// JUMA-TRX2 Rotary Encoder routines
// Juha Niinikoski OH2NLT 24.01.2006

// Microchip C30 version

// Trigger also from encoder phase B added 26.11.2006
// Encoder coarse read added 04.12.2006
// Very fine back and forth tuning anomaly corrected 21.12.2006
// Changed to every edge sensitive 26.12.2006 

// Includes
#include "juma-trx2.h"

// External Functions

// Rotary encoder
int enc_accu;						// Encoder pulse accumulator
unsigned char enc_event;			// Encoder event happened
static int last_phase;				// Last processed phase, A = 0, B = 1

// Encoder IRQ routines
// Very fast IQ encoder handler, no debounce, needs clean signal edges

// Encoder Phase A / RA15 / INT4
void __attribute__((interrupt)) _INT4Interrupt(void)
	{
	_INT4EP = !_INT4EP;				// Change sensitive edge
	_INT4IF = 0;
	enc_event = 1;

	if(last_phase != 0)				// Decode only if previous event was phase B
		{
		if(_INT4EP == 0)			// Rising edge action
			{
			if(ENQ_B)
				enc_accu++;
			else
				enc_accu--;
			}
		else						// Falling edge action
			{
			if(ENQ_B)
				enc_accu--;
			else
				enc_accu++;
			}
		}

	last_phase = 0;					// Mark phase A
	}

// Encoder Phase B / RA14 / INT3
void __attribute__((interrupt)) _INT3Interrupt(void)
	{
	_INT3EP = !_INT3EP;				// Invert trigger polarity
	_INT3IF = 0;
	enc_event = 1;

	if(last_phase == 0)				// Decode only if previous event was phase A
		{
		if(_INT3EP == 0)			// Rising edge action
			{
			if(ENQ_A)
				enc_accu--;
			else
				enc_accu++;
			}
		else						// Falling edge action
			{
			if(ENQ_A)
				enc_accu++;
			else
				enc_accu--;
			}
		}

	last_phase = 1;					// Mark phase B
	}

// Rotary Encoder system init
void encoder_init(void)
	{
//	ICP9 = 0x0010;					// Set priority to 1, default = 4
	_INT4IF = 0;
	_INT4IE = 1;
	_INT3IF = 0;
	_INT3IE = 1;			
	}

void reset_encoder(void)
	{
	enc_accu = 0;
	enc_event = 0;
	}

int encoder_get(void)				// Return with encoder reading
	{
	int x;

	x = enc_accu;
	reset_encoder();
	return(x);
	}

int encoder_get_coarse(int limit)	// Return with +/-1 if encoder is moved beyond limit.
	{
	if(enc_accu > limit)
		{
		reset_encoder();
		return(1);
		}

	if(enc_accu < -limit)
		{
		reset_encoder();
		return(-1);
		}

	return(0);						// Return with 0 if not moved
	}

