// JUMA-TRX2 Receiver Controller & DDS
// Juha Niinikoski, OH2NLT 22.11.2006

// MPLAB C30 version 2.05.6

// MCU dsPIC30F6014A
/*
 Project started 22.11.2006
 User interface & calibration displays 01.12.2006
 Split & RIT logic 02.12.2006
 Main board 74HC595 SPI interface added 03.11.2006
 Keyer code added 08.12.2006
 Keyer operating mode select added 10.12.2006
 Power off button timer & factory default load 12.12.2006
 Button tones 13.12.2006
 SWR calculations 19.12.2006
 All sorts of user interface modifications, EEPROM checksum 26.12.2006
 Service mode separated to different module 27.12.2006
 SWR alarm, fwd voltage bar graph, tune start beep 30.12.2006
 Yaesu CAT emulator, first test version 12.01.2007
 VFO lock function added 18.01.2007
 Iambic mode B added to keyer code 27.01.2007
 EEPROM structure separated to own trx2_eeprom.h file 28.01.2007
 ISD17240 Voice memory code tests 11.03.2007
 External keyboard module added 12.03.2007
 ISD17240 Voice memory installed flag, more memory banks & user interface changes 17.03.2007
 ISD17240 Voice memory Rev.B board definition changes 21.03.2007
 S-meter scale improvement. S9 marker and end stopper added 03.04.2007
 Squelch code 08.04.2007
 All-Band PA filter select logic & user interface tones improved 15.04.2007
 PA RF attenuator test 23.04.2007
 PA RF filter & attenuator selector code 17.05.2007
 MODE stored separately for both VFO's 17.05.2005
 PA RF attenuator (Power level) bits to EEPROM save 02.06.2007
 Different upper frequency limit for normal and service mode 03.06.2007
 MODE and VFO changes disabled when TX on, when Split selected Keyer and Tune operation corrected 06.06.2007
 Version changed to 1.00, first release package 19.06.2007
 RS232 test loop added 09.07.2007, v1.01 

 Interrupt driven RS232 RX driver & baud rate select 22.01.2008, test version
 Some Service mode displays changed
 Mute test function corrected
 Yaesu CAT module freq calculations fine tuned 23.01.2008, test version 1.02 
 CW TX side tone SCAF setup added 25.01.2008
 Yaesu CAT set frequency VFO select corrected, CW shift adjustment added 31.01.2008
 Compiler bug work around in uart.c module 01.02.2008
 CW pitch adjustment 300 - 1100Hz added 01.02.2008
 AUX UART2 interface & AUX Voice memory control added, leave UART1 for other use 02.02.2008 
 PWR display peak hold logic 07.02.2008
 RIT display changed 08.02.2008
 Yaesu CAT set operating mode command vfo indexing corrected 09.02.2008, v1.02
 DDS test code 06.03.2008
 Number of VFO memory banks increased 08.03.2008, v1.03 test version
 Startup changed to show version instead of date 11.03.2008, v1.03
 Bug fixes, Split mode CW keyer problem, CW left on if VFO pressed while keying  03.07.2008, v1.04 Test
 LCD back light timer code added 03.07.2008 v1.04
 TRX2 command protocol 14.08.2008 v1.05
 More TRX2 protocol commands 17.08.2008 v1.05
 UART1 init sequence changed to be compatible with later silicon 30.10.2008 v1.05
 In serial test.c bar graph test 'B' changed to 'b', better PA100 protocol compatibility  15.11.2008 v1.05
 TRX2 protocol bug fixed, own timer (pa100_tx_tmr) added 15.11.2008 v1.05
 TRX2 protocol capture start condition added 16.11.2008 v1.05
 UART module data type cast corrections 18.11.2008 v1.05
 Yaesu CAT Read TX status command 0xF7 bug corrected, squelch bit added to 0xE7 response 28.01.2009 v1.06 test
 Undocumented feature discovered from Yaesu 897 CAT protocol
 0xE7 response is 0xFF when TX and 0xF7 response is 0xFF when RX 29.01.2009 v1.06 test
 Dirty patch for undocumented Yaesu commands 00 xx 00 00 BB to get HRD working 30.01.2009 v1.06 test
 TX / PTT control added to JUMA-TRX2 protocol 17.04.2009 v1.06 test
 K250 & K600 filter switch points changed 18.04.2009 v1.06 test
 Split mode VFO copy strategy changed. Visible VFO is copied to another 18.04.2009 v1.06 test
 Memory can be selected keeping VFO button down and turning encoder knob 19.04.2009 v1.06
 DEFAULT_TX filter changed to 2,6kHz (2586Hz), 19.04.2009 v1.06

 ---------------------------------------- 5B4AIY Modifications ----------------------------------------

 Changed Brown-Out voltage from 2.0V to 2.7V. In v3.30b of the compiler, the file p30F6014A.inc now reserves
 the 2 volt value. - 5B4AIY
 Added Rapid band Switch feature - 5B4AIY
 Added additional memories for User Amateur Band frequency storage - 5B4AIY
 Added Help feature to serial test and fixed two bugs - 5B4AIY
 Added out-of-band frequency detection and transmitter inhibit - 5B4AIY
 Added auto sideband select - 5B4AIY
 Added an additional tuning rate of 1Hz, as well as user selectable frequency display format - 5B4AIY
 Changed DDS synthesiser calculation to be in decimal - 5B4AIY
 Added out-of-band annunciator to main display - 5B4AIY
 Greatly improved the accuracy of the measurement system - 5B4AIY
 Added user selectable alarms - 5B4AIY
 Greatly optimised the code - 5B4AIY
 Added VFO frequency rounding - 5B4AIY
 Added notification of mode/filter changes to all display screens - 5B4AIY
 Added Auto-Equalise to Serial Test Suite - 5B4AIY
 Added vfo copy feature - 5B4AIY
 Added extended commands to TRX-2 protocol so that the TRX-Manager program could work with this transceiver using
 a native command set rather than an emulation. - 5B4AIY
 Added CWR mode and completely changed how the CW modes work. - 5B4AIY
 Added prompted exit to menus and power down - 5B4AIY
 Corrected a minor error in the out-of-band detection. The flag was set when the carrier was actually on the band
 edge rather than outside the band. Thus, for example, on 20m 14.000000MHz was considered out of band rather than
 13.999999MHz. Similarly, 14.350000MHz was considered out of band rather than 14.350001MHz. This has now been fixed
 by a trivial change in the check_freq() function. - 5B4AIY A. Ryan 30/MAY/2012
 Collected all the User Configuration selections and System Setup code into individual small functions, and then assembled
 into two arrays. The configuration select code greatly simplified as a result. - 5B4AIY
 Modified out-of-band annunciator to blink instead of remain steady - 5B4AIY
 Modified Band Limits to account for USA General Class as well as Extra Class - 5B4AIY
 Relocated Band Limits adjustment from User Configuration to System Setup, this will cause a checksum error on
 initial startup as the mapping will have changed. - 5B4AIY
 Fixed minor bug. I had not appreciated that US amateurs are allowed to use all of the 160m segment down to 1800kHz - 5B4AIY
 Fixed trivial bug in Serial Test Suite 'b' command. This test confirms that the character generator RAM in the LCD has been
 loaded with the correct data for the bar graphs, as well as displaying the complete character set. It uses sprintf() to
 format a string which is placed in the display buffer. The function lcd_putst() is then passed the address of the buffer
 and it will display the string. If the character to display is a NULL byte, then since sprintf() merely copies the
 character to the buffer, when lcd_putst() attempts to display the buffer contents it discovers this NULL byte which it
 correctly interprets as the string terminator. This has now been fixed, and for a NULL byte the character is changed to
 a space, since the NULL byte is a non-displayable character anyway. Adrian Ryan - 5B4AIY 23/JUN/12
 Finally fixed out-of-band bug. See comments at line 3415. Adrian Ryan - 5B4AIY - 24/JUN/12
 Modified the Yaesu CAT code. Saved 594 bytes by optimising the 03 command substituting bin2bcd() and bcd2bin() functions
 instead of the inline code. Adrian Ryan - 5B4AIY - 25/JUN/12
 Also inserted a 'special' command that can be used with the LDG-817 Auto-Tuner. It is commented-out in this version.
 Tidied up the code in dm_0() and dm_3() so that the space or stopper bar character is now directly displayed. A.Ryan - 5B4AIY - 05/JUL/12
 Tidied up some code in lcd-trx2.c A.Ryan - 5B4AIY - 06/JUL/12
 Removed an unnecessary test in the disp_rit() function in lcd-trx2.c A.Ryan - 5B4AIY - 08/JUL/12
 Note: lcd_spc() is a macro definition in juma_trx2.h A. Ryan 31/JUL/2012
 Added command to serial_Y_cat.c to permit operation with LDG Z-817 autotuner. A.Ryan - 5B4AIY - 28/AUG/2012
 Added error messages to trap handler. A. Ryan - 5B4AIY - 12/SEP/12
 Cleaned up some code by making variables static instead of global. A.Ryan - 5B4AIY - 25/SEP/2012
 Revised S-Meter scale calculation to simplify it and remove excessive precision.
 The previous S-Meter calibration constant can be converted to the new value as follows:
 New Calibration Factor = 65536 / Old Factor. For the original transceiver using the ST112 transistor, the
 default calibration factor was 1920, its new value would be: 65536/1920 = 34.
 With the new transistor, my calibration factor was 3900, thus my new factor is 65536/3900 = 17.
 Changed SWR alarm behaviour. If in the TUNE mode, then High SWR alarm is delayed by an adjustable time to allow
 you to tune an antenna manually. The time limits are from 1 to 10 seconds, or disabled. If the delay is disabled,
 then the actual delay reverts to the normal delay time of 500mS.
 Changed SWR display. The pseudo-analog graphic meter now displays SWR as a fraction of the trip limit.
 Added limits displays. The pseudo-analog now displays voltage as a suppressed-zero voltmeter, the lower scale
 limit being the under-voltage trip, full-scale is the over-voltage trip. The ammeter is displayed as a fraction
 of your over-current trip. These limits displays are only visible in the transmit mode, in receive the S-Meter
 is displayed. The RF power indication now shows true power, and the scale is 10W maximum although the digital
 display will still indicate powers in excess of this.
 The mapping of the EEPROM has been rationalised. The calibration settings for id_mult, batt_mult, and s_mtr_cal were
 changed from long to int. A.Ryan - 5B4AIY - 21/OCT/2012
 Optimised and rationalised the adjustments. The encoder sensitivity is now set to normal, medium or slow depending
 upon the parameter being adjusted. For example, the backlighting and contrast adjustments were far too fast, and
 have now been slowed down. The adjustments for other parameters are similarly set to more reasonable values
 especially the On/Off adjustments or others where the actual values are either zero or one.
 Fixed a minor low-level bug. In the Calibration Menu, the mode was accidentally set to CWR, it has now been
 corrected to the TUNE mode.
 Fixed a minor bug. In the Calibration Menu, the default values would only be saved if an explicit save operation
 was performed rather than being restored and save automatically. This is now fixed. If there is a checksum
 both the calibration and the configuration values are restored and saved. A.Ryan - 5B4AIY - 26/OCT/2012
 Fixed minor bug. In the Serial Test Suite the value of the long SWR delay time was not correctly displayed.
 Optimised the serial_test() function by collecting some common code into a function. A.Ryan - 5B4AIY - 11/NOV/2012
 Added fast attack/slow release peak power display to complement the same display in the linear amplifier. A.Ryan - 5B4AIY - 20/NOV/2012
 Build 12. Fixed minor bug in default settings. The Long SWR Delay was accidentally set to a huge value.
 This has now been fixed, and the default value of 5 seconds is used.
 Slight re-arrangement of order of User Configuration settings. The S-Meter Hold Time setting now follows the
 S-Meter Selection setting. A.Ryan 24/NOV/2012

 V1.09a
 Build 1 Changed CW Release Time from 7 * CW Period to an adjustable time from 200mS to 1,500mS for Laurent Labourie.
 Note: This only applies to the CW Straight mode! A.Ryan - 5B4AIY - 26/NOV/2012
 Build 2
 Replaced a number of single-line functions with macros. (See juma-trx2.h) A.Ryan - 5B4AIY - 30/NOV/2012
 Modified the User Configuration prompts to be rather more attractive.
 Altered the display contrast to have a minimum value of 50.
 Moved function display_out_of_band() to lcd-trx2.c
 Set the minimum CW Release time to 200mS. A.Ryan - 5B4AIY - 30/NOV/2012
 Build 3
 The CW Release Delay page is now only visible if the keyer mode is set to STRAIGHT. A.Ryan - 5B4AIY - 2/DEC/2012
 Build 4
 Moved the CW Release Delay menu item so that it follows the Keyer Mode selection. A.Ryan - 5B4AIY - 3/DEC/2012

 V1.09b
 Build 1
 Fixed very minor display bug. When using the multi-memory VFO option, the lock and tuning rate status was lost
 when using the fast memory select with the VFO knob, and was restored when the VFO button was released.
 The lock status of each VFO and the current tuning rate are now continuously displayed.
 Fixed subtle bug in the Extended TRX-2 Protocol. If the Extended Mode command was used, it would respond the
 first time, but then ignore subsequent commands. A.Ryan - 5B4AIY - 05/JAN/2013
 Build 2
 Removed tuning rate from multi-memory status display. The tuning rate does not change, and as the status
 is already written to the display memory, there is no need to constantly over-write it. This saved 18 bytes of
 code that was simply address calculation and character retrieval. A.Ryan - 5B4AIY - 06/JAN/2013
 Build 3
 Simplified the default restoral in the service menu. A long push is no longer required, instead
 a normal prompted exit is given. This also allowed for a simplification of the default restoral via the MODE
 button when powering up. The save_settings() function was modified to add another button test to the PWR button
 test to permit the removal of a number of lines of redundant code.
 Added a global variable - RS_232_Mode - a pointer to a function taking void and returning void, whose value
 is set once during initialisation or if the User Configuration variable: eeprom.defval.serial_mode is changed.
 This avoids a needless loop invariant address calculation in the main loop.
 Removed redundant test when checking the keyer mode to skip a display page.
 Moved some system status flags initialisations to the declaration section.
 Moved the setting of split_flag from lcd_0() to the start of the main loop, this permitted a great simplification
 of the Split Mode logic, especially the setting of the cw_active flag. A.Ryan - 5B4AIY - 14/JAN/2013
 Build 4
 The display LCD code has been moved to the top of the main loop. This prevents a spurious, very very brief
 display of 'VFOC' when changing from VFO-B to the Split mode.
 The DDS load and SPI bus activate code moved to the end of the main loop so that it executes after all the
 frequency and mode changes have been made. A.Ryan - 5B4AIY - 15/JAN/2013
 Build 5
 Relocated all the prompts from service.c, isd1700.c, and juma-trx2.c to the data segment. The prompts in
 serial_test.c are too numerous to bother with. A.Ryan 16/JAN/2013
 Build 6
 Put all the prompts from service.c, isd1700.c, and juma-trx2.c back where they were. The relocation had no
 effect on the cycle time of the main loop. Corrected declaration of pointer: *user_band_prompt so that it
 points to const chars, thus avoiding a compiler warning. Corrected declaration of putst2() argument (uart2.c)
 so that it points to const chars avoiding the same compiler warning. Improved the construction of
 putst2() so that it is more concise. A.Ryan - 5B4AIY - 17/JAN/2013
 Build 7
 Reduced the length of lcdpbuff[] from 40 bytes to 20 bytes. This buffer is exclusively used by sprintf()
 to format LCD display strings. All the uses utilise defined length strings, and thus there is no possibility
 of over-running the buffer, even 20 bytes gives us a safety margin.
 Added const qualifier to standard format messages in serial_trx2.c 
 Removed function: set_cgram() from lcd-trx2.c - never referenced. Optimised set_ch_bits()
 Fixed minor bug in Serial Test. The Bar Graph test would not display an end-of-bar symbol if the
 S-Meter was set to numeric. This is now fixed. A.Ryan - 5B4AIY - 23/JAN/2013
 Build 8
 Optimised auto_equalise() removing some redundant code and a function only referenced once.
 Minor optimisation of adjust_rf_attenuator() Removed find_rf_filt_idx() function and incorporated it
 as in-line code in send_packet_to_74hc595(). It was only ever called by this function.
 Collected literal text strings in isd1700.c to constant character arrays.
 Removed init_spi2() in isd1700.c, and made it in-line code. It was only ever called once.
 Collected literal text strings in serial_ext_kb.c to constant character arrays.
 Collected literal text strings in serial_trx2.c to constant character arrays.
 Eliminated send_cat() in serial_Y_cat.c, and made it in-line code. Only called once.
 Optimised ms_delay() in tmr5delay.c by more concise construction.
 Optimised clr_tx_buffer() and clr_rx_buffer() functions in Kenwood_CAT.c Modified RIT frequency
 format descriptor in Kenwood CAT module to remove redundant + correction. A.Ryan - 5B4AIY - 24/JAN/2013
 Build 9
 Revised the Kenwood CAT logic. Removed a lot of redundant code and simplified the split logic.
 Introduced the select_state() function to set the state of the transceiver from the commanded transmit
 and receive VFO selection. A.Ryan - 5B4AIY - 26/JAN/2013

 v1.09c
 Build 1
 Extended the command set of the Yaesu, Kenwood, and TRX-2 serial protocols.
 The Yaesu emulation now contains the F8 Extended Display, E0 Tuning Rate adjustment, and the PKT mode
 in the mode command which turns the TUNE mode on. (See serial_Y_cat.c for details)
 The Kenwood emulation now contains the XD Extended Display, and has also been extensively re-written.
 The tx_buffer[] has been removed, along with the clr_tx_buffer() function, no longer required,
 The TRX-2 protocol is enhanced with the Y Extended Display. This module has been significantly
 optimised. Removed the set_FREQ() from the serial_trx2.c and made it in-line. Only called once.
 Re-wrote the check_freq() function to eliminate the switch-case and perform the General Class band
 checks in a single go. This entailed adding some dummy values to the lower_segment_limit[] and
 upper_segment_limit[] arrays, but saved a significant amount of code. A.Ryan - 5B4AIY - 27/JAN/2013
 Build 2
 Fixed inoperative VFO lock bug in Yaesu CAT module. Fixed bug in Auto-Inform. If the latency was set
 to something other than 1, then the VFO mode would not automatically be changed to A+B (Split).
 Removed filter_nop_char() in serial_trx2.c and replaced with simplified in-line code. Only called once.
 Removed several calls to clr_buffer() in serial_trx2() and used single call at end of composite switch-case.
 Simplified the get_num() in serial_trx2.c. Since only a single ASCII character is ever being converted,
 it is simpler to calculate the number by subtracting the character '0' from the byte. The limit has been
 modified to eliminate the unnecessary +1 addition in the limit check by simply increasing the
 value in each of the called instances. Fixed minor bug in frequency reporting, the '=' symbol was missing.
 A.Ryan - 5B4AIY - 30/JAN/2013
 Build 3
 Fixed minor bug. During the boot sequence, the Voice Memory message is sent out before the serial port has
 been set to its final speed. This message has now been removed from the isd1700.c module to the juma-trx2.c
 module, and the boot messages section considerably simplified and improved. A.Ryan - 5B4AIY - 6/FEB/2013
 Build 4
 Modified the serial_Y_cat.c. The clear_cat() now returns the command index as zero. The initialisation
 code has been considerably simplified. Unnecessary code to clear the buffer removed. A.Ryan - 5B4AIY - 10/FEB/2013
 Build 5
 Initialised variable: low_voltage_warning to zero, and removed unnecessary code.
 Removed variable: ultemp, only ever used once, replaced with in-line code. A.Ryan - 5B4AIY - 16/FEB/2013
 Build 6
 Added a fully decoded SPI bus data display to the Serial Test Suite to replace the existing display. 19/FEB/2013
 Build 7
 Removed redundant double variable v in service.c and modified W_Fmt message in lcd-trx2.c
 Re-organised timers_pwm.c. Set maximum backlight to 1000 in juma-trx2.h. Set initial backlight
 and contrast levels to their default values in the initialisation routine. A.Ryan - 5B4AIY - 04/MAR/2013
 Build 8
 Fixed minor bug. When using the backlight timer option, the backlighting was not turned on if an alarm occured.
 This has now been fixed. A.Ryan - 5B4AIY - 22/MAY/2013

 v1.10a
 Build 1
 Modified the Squelch. In the previous firmware if the backlight timer and squelch features were used, if a signal
 exceeded the squelch threshold the backlight was not turned on. There are now two modes. Squelch-A operates as before,
 Squelch-B turns the backlight on if the received signal is above the squelch threshold. A.Ryan - 5B4AIY - 25/MAY/2013
 Build 2
 The backlight timer was previously adjustable in 1 second increments. I seriously doubt that anyone ever used an
 interval of less than 5 seconds. The timer now increments in 5-second intervals. A.Ryan - 5B4AIY - 27/MAY/2013

 The accuracy of the current measurement could be improved by incorporating the idle and backlighting currents with
 the measured PA current. The principle is based upon measurements made of the currents drawn by the transceiver
 in the receive and transmit states, with the backlighting on and off and at various settings, as well as on
 the various amateur bands. In the receive state, the lowest current occurs on 160m and with the backlighting off.
 The currents are essentially constant over the 11.0V - 15.0V input voltage range, and comprise three components:

 1. The fixed idle current, measured as 380mA
 2. The variable backlighting current, 10mA at a backlighting setting of 50 (minimum) up to 70mA at 1000 (maximum)
 3. The increase in current as the various amateur bands are engaged:
	Band	Current
	160m	0
	80m		10mA
	40m		20mA
	30m		25mA
	20m		35mA
	17m		40mA
	15m		50mA
	12m		55mA
	10m		60mA

  Since there is already a variable, band_in_use that registers the current band, then by declaring an array of values
  these can be indexed by band_in_use to arrive at the band current factor:
  
  const char band_current[] = { 0, 0, 10, 20, 25, 35, 40, 50, 55, 60 };
  Ib = band_current[band_in_use]
 
  Although the current increases as the audio output increases, as there is no connection to the microprocessor there
  is no real way that this can be accomodated.
 
  In the transmit mode, the current comprises the previous idle current and band current, but in addition, the bias
  current of the driver and amplifier stages which vary with supply voltage. With the backlighting set to minimum the
  measured transmit currents were:
 
	11.0V 620mA
	14.7V 760mA

  Since the transmit driver stage and amplifier current is actually measured, it is only necessary to add the previous
  idle current and band current to the measured driver and amplifier current to arrive at the total current.

  Thus, in receive, if the idle current is Ic, and the backlighting current Il, and the band factor Ib, then
  the displayed current is: Ic + Il + Ib
  If the backlighting is off, the Il term is set to zero.
  Ib is derived from either an array entry or from a calculation related to the VFO frequency.
  The backlighting current is derived from a simple ratiometric calculation based on the backlighting intensity setting.
  
  Il = 7 + 60 * Bl / 950 mA
  
  Thus, the display current is:
  
  Id	= 0.38 + Il + Ib
 
  and this current is also added to the current measured from the driver/amplifier board in transmit.
 Build 3
 Added display_screen() function to optimise displays, and changed format descriptor for s9_msg to eliminate
 uneccessary spaces.

 1.10b
 Build 1
 Optimised the power down sequence. The counted down sequence has been removed, as it was completely redundant.
 The new sequence has a choice of Prompted power down, where you can choose to save the current configuration,
 or an immediate power off without saving anything, or an immdiate power off and where the configuration is
 always saved. A.Ryan - 5B4AIY - 03/JUN/2013
 Build 2
 Fixed minor bug. If the "Save State" power-off sequency was chosen, the saved message was not sent to the serial
 port if the TEST mode had been selected. A.Ryan - 5B4AIY - 04/JUN/2013

 1.10c
 Build 1
 Fixes bug in Service Menu. The measurement for the drain current was not being made. Also altered the drain current
 display to have a peak reading long hold ammeter rather than a running average. In the Service Mode, the peak hold
 is inhibited, and an instantaneous display is used. A.Ryan - 5B4AIY - 22/JUN/2013
 Build 2
 Removed duplicate code and used the get_pwr() function instead. Set the TUNE attenuation to -1dB in the auto-equalise
 mode to get a more accurate representation of the PA gain variation. Modified the display_s_meter() function. It
 now takes an integer parameter describing the S-Meter type, 0 = Graphic, 1 = Numeric. A.Ryan - 5B4AIY - 23/JUN/2013

 1.10d
 Build 1
 Every other amateur transceiver that I know although permitting split frequency operation, does not permit the frequencies
 to be on different bands. This version of the firmware fixes this anomaly, so that when setting the transceiver
 to the 'A/B + Split' mode if the Rapid Bandswitch facility is used then the frequencies are checked and if the
 A and B VFOs are not set to the same amateur band, then if VFO-A is the active VFO, VFO-B is forced to the same frequency
 and mode. Similarly if VFO-B is the active VFO, then VFO-A is forced to the same frequency/mode. If someone actually
 has a use for cross-band operation whereby you use VFO-B as the transmit VFO, and VFO-A as the receive, but they
 are tuned to different bands, then this is still possible if you copy a frequency from one of the VFO memories
 into the active VFO. A.Ryan - 5B4AIY - 17/JUL/2013 - Happy Birthday, Emma!
 Build 2
 Replaced select_nearest_band() with check_freq() when checking split mode Rapid Band-Switch frequency assignments.
 This eliminates the possibility of a non-adjustment when the inactive VFO frequency is only 'near' the amateur band,
 rather than being in-band. A.Ryan - 5B4AIY - 18/JUL/2013
 Build 3
 Cleaned up some code in timers_pwm.c associated with the auto-keyer by adding the function _cw_active(int). A.Ryan - 5B4AIY - 04/AUG/2013

 1.10e
 Build 1
 Added Vibroplex mode to the CW feature set. In this mode the keyer will produce automatic dots, but only manual dashes,
 emulating the famous Vibroplex 'bug' key. This mode is primarily designed for a single paddle key. If a dual-paddle key
 is connected, then squeezing both paddles causes the symbol to stop. A.Ryan - 5B4AIY - 05/AUG/2013
 Build 2
 Optimised the keyer logic by only reading the state of the DOT/DASH I/O ports once and assigning the values to dot and
 dash local variables. A.Ryan - 5B4AIY - 05/AUG/2013
 Build 3
 Optimised the play dot code in the Vibroplex keyer logic to remove a redundant test and use the _play_dot() funtion
 instead. A.Ryan - 5B4AIY - 07/AUG/2013
 Build 4
 Removed the unnecessary duplication of the cancel key for the prompted exit. Only the PWR button now aborts the save
 operation, the FILTER button has been eliminated. A.Ryan - 5B4AIY - 09/AUG/2013
 Build 5
 Modified the RSR-232 mode selection logic so that the Voice Memory is only selectable if the optional module is
 actually present. A.Ryan 5B4AIY - 17/AUG/2013

 1.10f
 Build 1
 optimised the Voice Memory Command functions. In the previous versions, both UART1 and UART2 when used as the VM
 control ports had essentially the same command parsers. Since there should never be a situation where both UARTs
 are used to control the VM, this duplication of code is uneccessary, and wasteful. The command parser was implemented
 as the _VM_Command() function, and utilised by both the serial_ext_kb() and uart2_ext_kb() functions. Appropriate logic
 was inserted into the uart2_ext_kb() function to ensure that it would only control the VM if UART1 was not set to the
 Voice Memory serial mode.
 Build 2
 Removed redundant "Voice Memory Not Installed" message and test. With v1.10e Build 5 if the Voice Memory is not installed,
 then this message would never print, since UART1 could never be set to the Voice Memory mode if the module was not installed.
 In the uart2_ext_kb() function, there is now a test, and if the VM is not installed, then the function simply returns.
 In addition, removed all the UART1 Voice Memory messages. These are redundant as the LCD displays what is happening,
 and if triggered by UART2 commands, would simply confuse any terminal activity associated port 1. In any case, when
 using either the external keyboard or the Windows VM Controller program, none of the messages would be displayed.
 These changes saved 879 bytes of memory. A.Ryan - 5B4AIY - 18/AUG/2013
 Build 3
 Completely removed the uart2_ext_kb.c file. After the previous changes, the only thing left in this file was the
 uart2_ext_kb() function. The code for this function has now been moved into the serial_ext_kb.c file where it more
 properly belongs. A.Ryan - 5B4AIY - 19/AUG/2013

 1.10g
 Build 1
 Optimises a lot of functions in isd1700.c Collected common code and made into functions: _disp_meter(), and
 check_stop(). Fixed display bug. Fixed bug where graphic S-meter and Power meter scales were not being fully displayed.
 When the S-Meter was set to NUMERIC, the end-of-scale bar was missing when transmitting from Voice Memory - fixed.
 Fixed bug where numeric S-meter was always displayed when recording from receiver audio. Fixed out-of-band transmit
 from the voice memory bug. A.Ryan - 5B4AIY - 20/AUG/2013
 Build 2
 More optimisation of isd1700.c Removed the display_time() function, it was only called once, incorporated the code
 in the calling function, display_meter(). Rationalised the display of the record from RX audio. Added code to the
 check_stop() function so that the serial buffers are flushed. Fixed minor bug. On switch-on, very briefly, the first time
 through the main loop, the transceiver is in the TX mode until the I/O status is correctly read. Presetting PTT_IN to the RX
 mode in the declaration section fixes this. Relocated the UART1/UART2 buffer flush to be immediately after initialising
 the UARTs. Fixed bug related to squelch settings. Added code to turn backlight on when using the Voice Memory, and checks
 so that if the mode is not LSB/USB, as well as if the frequency is out-of-band then an error message is displayed.
 A.Ryan - 5B4AIY - 22/AUG/2013
 Build 3
 More Voice Memory optimisation. Modified tx_rx_16_spi2() and tx_rx_56_spi2() so that they no longer return any data.
 In all previous versions the return data was discarded anyway. Modified tx_rx_32_spi2() so that the returned data
 is masked to 8 bits. The return data is only ever used during the device initialisation to determine the device ID
 when called by init_isd1700(). Changed operation so that when recording from receiver audio, the memory is preset to
 M0, and the command is immediately executed. Added code to display error messages if the transceiver is not in a speech
 mode (LSB/USB) or if the frequency is out-of-band if the Transmit command is issued. Removed two spurious memory address
 setup commands. A.Ryan - 5B4AIY - 23/AUG/2013
 Build 4
 Changed the elapsed time display of the Voice Memory. In a RECORD mode, the time now decrements to zero. In a REPLAY
 mode, the time increments to either the maximum memory value, or the actual message recording time, whichever occurs
 first. Memory 0 is 120 seconds, Memory 1 is 20 seconds, Memories 2 through 8 are 10 seconds, and Memory 9 is 14 seconds.
 Note that as the Voice Memory chip clock is not synchronised with the main system's clock, and further, as it is generated
 by means of a resistor, the tolerances associated with both the chip's timing logic as well as that of the resistor means
 that the memory slot times are only an approximation. A.Ryan - 5B4AIY - 11/SEP/2013
 Build 5
 Added termination prompt to UART1/2 cross-copy test in serial_test.c and collected the common code for the PA Filter
 attenuator test into a single module. A.Ryan - 5B4AIY - 14/SEP/2013
 Build 6
 Added array slot_time[] in isd1700.c and used this to calculate the slot times to display.
 This saved 39 bytes. A.Ryan - 5B4AIY - 25/SEP/2013
 Build 7
 Modified check_freq() so that only a single parameter is passed, and the second frequency parameter is derived in the function.
 This was done because the original version passed two frequency parameters, the upper and lower frequency, but they were always
 identical, so it made more sense to only pass one as this reduced the amount of complex address calculation. This modification
 saved 15 bytes. A.Ryan - 5B4AIY - 30/OCT/2013
 Build 8
 Modified get_one_zero(), get_value(), and set_decade_value() so that they are now passed a pointer to the values they modify
 rather than the value itself. This avoids a lot of duplicate address re-calculation for the configuration and calibration memory
 values, and saved 60 bytes. A.Ryan - 5B4AIY - 20/NOV/2013
 Build 9
 Modified Service Mode. The save settings logic was somewhat convoluted. Now, the FAST/VFAST switch is tested first, then the
 check for the maximum or normal page is made. This eliminated some redundant logic tests and saved 82 bytes. A.Ryan - 5B4AIY - 3/DEC/2013
 
 1.10h
 Build 1
 Added Fast Page Switch feature to Service & Calibration mode. Fixed minor bug. The service mode print-out if the serial
 mode was TEST was not occuring - fixed. A.Ryan - 5B4AIY - 03/DEC/2013
 Build 2
 Modified User Configuration and Service beeps so that an acknowledgement beep is always given in these modes, even when the
 beep length is set to zero. A.Ryan - 5B4AIY - 04/DEC/2013
 Build 3
 No code changes. Discovered that the contents of file: lcd-trx2.h was duplicated in file: lcd-trx2.c, as well as being used
 in other files. Removed spurious re-definitions in lcd-trx2.c and included the header file instead. A.Ryan - 5B4AIY - 24/DEC/2013
 Build 4
 Modified the RIT_fmt[] descriptor to include the leading letter 'R', and also the disp_rit() function. A.Ryan - 5B4AIY - 24/DEC/2013
 Build 5
 Removed header file lcd-trx2.h and incorporated its contents into juma-trx2.h. A.Ryan - 5B4AIY - 27/DEC/2013
 Build 6
 Removed set_cur_lcd() function, and replaced it with lcd_cmd(). Changed all the cursor positioning commands
 so that the intrinsic OR operation was already implemented. Optimised initlcd(). Changed old_data to be a
 static unsigned int in exchg_data_spi1() rather than static local. Removed local variable spi_rx_data_available
 never used. A.Ryan - 5B4AIY - 03/JAN/2014
 Build 7
 Optimised the SPI Interrupt routine. Added explanation of how the SPI bus operates in spi1.c. A.Ryan - 5B4AIY - 09/JAN/2014
 Build 8
 There is a redundant test in SPI interrupt. After a considerable amount of testing, it was obvious that the SPI1OV
 flag is always set after sending data to the SPI bus. Therefore, there is no need to test this flag, the dummy
 read operation is always required to clear the flag. However, the test is retained to preserve the latch pulse width.
 Modified the disp_meter() function for a more efficient form. This saved 123 bytes! A.Ryan - 5B4AIY - 10/JAN/2014
 Build 9
 Removed SWITCH-CASE construct for rapid bandswitch select and replaced with simpler IF construct. Removed a number
 of single-use temporary variables, and replaced with a common temporary variable. A.Ryan - 5B4AIY - 16/JAN/2014
 Build 10
 Added a few more manifest constants to juma-trx2.h Slightly changed the logic by which the TX Filter is recognised when
 exiting the User Configuration menu. In previous versions the filter was recognised by being > 2, now it is explicitly
 checked for equality to the constant TX_FILTER. A.Ryan - 5B4AIY - 23/JAN/2014

 1.10i
 Build 1
 Modified the Rapid Bandswitch feature to display its current mode. A.Ryan - 5B4AIY - 26/JAN/2014
 Build 2
 Modified set_range() in service.c so that order of minimum and maximum parameter values corresponds with those in
 get_value() for the sake of consistency. Modified SWR display to show 'SWR--.-' in receive and when no valid
 SWR measurement available. Modified max_min() so that it now takes a pointer to the value, and directly adjusts it
 via this pointer, returning void. A.Ryan - 5B4AIY - 28/JAN/2014
 Build 3
 Added the feature to save the current display page when saving the state of the transceiver. This makes use of the
 spare EEPROM location eeprom.defval.mode[39] which is an integer object, currently unused. A.Ryan - 5B4AIY - 31/JAN/2014
 Build 4
 Modified the previous feature to enable the user to select either the last selected page, or to have a selectable
 fixed default page displayed on start-up. A.Ryan - 5B4AIY - 01/FEB/2014
 Build 5
 Eliminated the complex SWITCH-CASE construct in display_settings() in favour of an indexed array. Added feature
 to allow selectable start-up display page or the last displayed page. A.Ryan - 5B4AIY - 2/FEB/2014
 Build 6
 Modified the prompts for the Start-Up Display Page feature, and added additional information to the Serial Test
 to display the actual last page that has bee saved. Removed the No/Yes prompt message, only used by the
 'Include RIT' setup. Used the existing Off/On message instead. A.Ryan - 5B4AIY - 4/FEB/2014
 Build 7
 Re-organised the order of the configuration pages so that the Power Off Mode and Display Page Select are next to each
 other. This is because the Display Page Select is now determined by the Power Off Mode. If the mode is IMMEDIATE then
 as no state save occurs, it makes no sense to allow the display page to be set to the Last Page. Thus the new logic
 checks the value of the Power Down mode, and if it is set to IMMEDIATE and the Display Page is set to Last Page, then
 the Display Page is reset to the default, MODE B/W, and the maximum allowed value for the display page is then
 PA CURRENT.
 Build 8
 Minor adjustment of prompt messages, and a major rationalisation of format descriptors for the User Configuration
 display pages.
 
 1.10j
 Build 1
 Finally fixed the spurious previous/next page select Fast Page Switch bug in the Calibration and
 User Configuration menus. A.Ryan - 5B4AIY - 08/FEB/2014
 Build 2
 Added SCAF_CLK definition to juma-trx2.h Since this constant is derived from the FCY constant, which is defined as an
 unsigned long integer, and the SCAF clock variable is a signed long integer, a type conversion occurs, which, whilst
 it does not alter the length of the code, does effect a minor change. This does not seem to affect the operation of the
 SCAF filter, so the actual clock frequency obviously ends up the same. A.Ryan - 5B4AIY - 09/FEB/2014
 Build 3
 Fixed User Configuration page skip bug. When moving the Power Down mode select, I omitted to change the page skip value.
 Removed redundant variables in several EEPROM save and restore functions. The return values were always discarded.
 Modified the way the Immediate power down mode operates. A flag has been added to show that a change has been made to
 either the User Frequency or VFO memories, and if this flag is set and the power down mode is Immediate, then you will
 now receive a last prompt when you power down to save the settings. A.Ryan - 5B4AIY - 25/FEB/2014
 Build 4
 Modified the way the Rapid Band Switch User frequenies and VFO frequencies are saved. In this version the save is now
 automatic, and is in the power_off() module. The previous IMMEDIATE mode power-down prompt is removed. A.Ryan - 5B4AIY - 26/FEB/2014
 Build 5
 Added a prompt to the Erase voice memory command. A.Ryan - 5B4AIY - 01/MAR/2014

 1.10k
 Build 1
 Modified the way the Press & Hold functions operate. All the P & H functions now latch the page so that you no longer need
 two hands to operate the Rapid Band Switch, VFO Memory Copy, Multi-Memory VFO Select, User Configuration Fast Page Select,
 or the Calibration Fast Page Select functions. With this version, press the appropriate button until the long beep sounds,
 release the button and the function will stay selected. Rotate the VFO knob to select the desired page/band/VFO/Memory, then
 cancel the page hold either by a brief press of the page select button originally used, or the PWR button. For the VFO select,
 User Configuration Fast Page Switch or Calibration Page Switch since these pages do not indicate any differently in the fast
 page switch mode, a rapid warning beep is given. For the Rapid Band Switch and VFO Memory Copy functions, since these pages
 intriniscally display their functions, no warning beep is given. A.Ryan - 5B4AIY - 03/MAR/2014
 Build 2
 For the sake of consistency, added the warning beep to both the Rapid Band Switch and VFO Memory Copy functions.
 Modified the adjust_rf_attenuator() function. It now only requires the selection of the TUNE mode, press the PTT switch,
 briefly press the PWR button, and the function is enagegd without having to hold the PWR button. A warning beep is also
 given to remind you that the function is latched. To save the new settings and exit, release the PTT switch.
 Changed the LONG_PUSH timer to 650mS, and the VERY_LONG_PUSH timer to 1,500mS. A.Ryan - 5B4AIY - 04/MAR/2014
 Build 3
 Replaced various temporary variables in main loop with temporary local variables, and removed the need to 'fix'
 spurious mode/filter/vfo changes. A.Ryan - 5B4AIY - 09/MAR/2014
 Build 4
 Fixed major bug. When using the W command in the TRX-2 protocol this wouldprovoke an Adress Trap Error! because the arguments
 passed to the max_min() were incorrect. Updated the J command. Updated the B command to use the current Default/User frequency
 settings. A.Ryan - 5B4AIY - 10/MAR/2014
*/
// Include files
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include "juma-trx2.h"						// JUMA-TRX2 hardware specific definitions
#include "DataEEPROM.h"						// EEPROM mapping
#include "trx2_eeprom.h"					// EEPROM storage structure

// Out-Of-Band Character
// (Not used in this version which blinks the decimal point when out-of-band Adrian Ryan - 5B4AIY 23/JUN/12)
//#define OUT_OF_BAND	0xA1					// Out-Of-Band character (Standard, 3x3 pixels, centre pixel blank
//#define OUT_OF_BAND	0xA5					// Out-Of-Band character (Alternative, smaller 2x2 pixels

// External Functions
extern void lcd_putch(unsigned char);       // Added - 5B4AIY
extern void ClearUSART1queue(void);			// Added - 5B4AIY
extern void ClearUSART2queue(void);			// Added - 5B4AIY
extern void clr_rx_buffer(void);			// Clear Kenwood RX buffer (Kenwood_CAT.c)
extern void clear_cat(void);				// Clear Yaesu CAT buffer
extern void clear_buffer(void);				// Clear TRX-2 PA-100D RX buffer (Kenwood_CAT.c)
extern void serial_kenwood(void);			// Kenwood CAT protocol handler (Kenwood_CAT.c)
extern void reset_state(void);				// Juma TRX-2/PA-100D TRX-Manager Protocol state initialisation (serial_trx2.c)
extern int get_ms_counter(void);			// General 1 second counter

extern void us_delay(unsigned int);			// Delay routines
extern void ms_delay(unsigned int);

extern void init_timers_pwm(void);			// Timers & PWM DAC
extern void init_scaf_clock(void);			// Initialise SCAF clocking
extern void set_scaf_clock(unsigned int);	// SCAF filter clock divider

extern void tone_on(int);					// Start tone generator, period length
extern void beep(int, int);					// Generate beep, period, length

extern void dds_load(unsigned long);
extern void dds_init(void);

extern void init_adc12(void);
extern int convert_adc12(unsigned int);		// ADC

extern void draw_s_meter(int);				// LCD functions (lcd-trx2.c)
extern void lcdoutch(unsigned char);
extern void lcd_putst(register const char *);
extern void set_ch_bits(int, int);
extern void set_chgen(int);
extern void initlcd(void);
extern void clear_lcd(void);
extern void lcd_cmd(unsigned char);
extern void disp_freq(long);
extern void disp_rit(int);
extern void disp_meter(double, int);		// New version - 5B4AIY (in lcd-trx2.c)

extern void InitUSART1(void);				// UART1 / Terminal Functions
extern void SetUSART1baud(int);				// Baud rate select
extern void putch(char);
extern int kbhit(void);
extern unsigned char getch(void);
extern unsigned long getlong(void);

extern void InitUSART2(void);				// UART2 / Secondary UART (Voice Memory Option)
extern void SetUSART2baud(int);	
extern void putch2(char);
extern int kbhit2(void);
extern unsigned char getch2(void);
extern unsigned char getche2(void);
extern void putst2(const char *);

extern void init_spi1(void);				// SPI, main board 74HC595 I/O latch
extern void exchg_data_spi1(unsigned int);

extern void encoder_init(void);				// VFO encoder functions
extern int encoder_get(void);
extern int encoder_get_coarse(int);
extern void reset_encoder(void);

extern void service(void);					// Service functions

extern void serial_test(void);				// Serial I/O
extern void dump_eeprom(void);				// Dump EEPROM content
extern void serial_yaesu(void);				// Yaesu CAT emulator
extern void serial_trx2(int);				// JUMA-TRX2 protocol
extern void serial_ext_kb(void);			// External keyboard (Voice Memory control)
extern void uart2_ext_kb(void);				// UART2 / AUX external kb handler

extern void init_isd1700(void);				// Init ISD1700 voice memory system

// External Variables
extern int idx;								// Used in Kenwood_CAT.c
extern int auto_inform;						// Used in serial_trx2.c
extern int A_equals_B_flag;					// Used in serial_trx2.c
extern int cw_period;						// Defines the period of a CW unit.
extern int decay_counter;					// Used to define the time the peak power is displayed.

extern unsigned char enc_event;				// VFO encoder moved flag.
extern unsigned int long_push;				// Long push timer for buttons.

// Processor Configuration
_FOSC(CSW_FSCM_OFF & EC);					// 30MHz clock = 7,5MHz cycle
_FWDT(WDT_OFF);                				// Turn off the Watch-Dog Timer.
//_FBORPOR(MCLR_EN & PWRT_64 & BORV_20);   	// Enable MCLR, power on timer, Brown out 2V
_FBORPOR(MCLR_EN & PWRT_64 & BORV_27);   	// Enable MCLR, power on timer, Brown out 2.7V - Changed to eliminate fatal compile time error - 5B4AIY
//_FGS(GEN_PROT);            				// Disable Code Protection, old compiler
_FBS(CODE_PROT_OFF);
_FSS(CODE_PROT_OFF);
_FGS(CODE_PROT_OFF);

// Global Variables
// ================================================================================================================
// Frequency decade steps for 180 MHz master clock
const long f_steps[] = {			// Indexed by eeprom.defval.step_idx
						1,			// 0 - 1Hz (Not Used)
						10,			// 1 - 10Hz
						100,		// 2 - 100Hz
						1000,		// 3 - 1kHz
						10000,		// 4 - 10kHz (Not Used)
						100000		// 5 - 100kHz (Not Used)
						};

const long rf_filt_attn[] = {		// Filter change points, used to set rf_filt_idx
							K250,	// 250kHz	0
							K600,	// 600kHz	1
							M1,		// 1MHz		2
							M2,		// 2MHz		3
							M4,		// 4MHz		4
							M8,		// 8MHz		5
							M12,	// 12MHz	6
							M15,	// 15MHz	7
							M19,	// 19MHz	8
							M23,	// 23MHz	9
							M26		// 26MHz	10
							};

const unsigned int rf_filter_bits[] = {			// Indexed by: rf_filt_idx
										0x0000,	// 0
										0x0100,	// 1
										0x0200,	// 2
										0x0300,	// 3
										0x0400,	// 4
										0x0500,	// 5
										0x0600,	// 6
										0x0600,	// 7
										0x0700,	// 8
										0x0700,	// 9
										0x0700,	// 10
										0x0700	// 11
										};

const char ver[] = {"v1.10k"};			// Software version information, displayed at startup
const char date[] = {"4 - 10/MAR/2014"};// Software build date information, displayed at startup
const char firmware[] = {"\n\r%sFirmware: %s Build: %s"};
const char copyright[] = {"\n\rCopyright: Juha Niinikoski - OH2NLT & Matti Hohtola - OH7SV\n\r"};
const char added_features[] = {"(Additional features and modifications - Adrian Ryan - 5B4AIY)\n\r"};

// Standard User Interface Displays
const char vfo_locked[] = {"VFO-%c is locked!"};
const char copy_vfo[] = {"Copy VFO%c %c VFO%c"};
const char out_of_band[] = {"  Out Of Band!  "};
const char not_in_user_mode[] = {"Not In User Mode"};
const char user_mem_locked[] = {"User Mem Locked!"};
const char selected_band[] = {"Band Select:%3im"};
const char save_to_band[] = {"Saved to:%6im"};
const char CW_rls_prompt[] = {"Delay:%10s"};
const char PWR_No_MODE_Yes[] = {"PWR:No  MODE:Yes"};
const char Display_Watts[] = {"  --.-W"};
const char Display_Amps[] = {"  -.--A"};
const char Emergency[] = {"   Emergency    "};
const char Shutdown[] = {"    Shutdown    "};
const char Loop_Back_Test[] = {"Serial Port Test"};
const char Awaiting_Data[] = {"Awaiting data..."};
const char Overcurrent_Alm[] = {"  Over Current  "};
const char High_SWR_Alm[] = {"    High SWR    "};
const char Over_Voltage_Alm[] = {"  Over Voltage  "};
const char Under_Voltage_Alm[] = {"  Low Voltage!  "};
const char Low_Battery_Alm[] = {"  Low Battery!  "};
const char Display_TX[] = {" TX"};
const char Display_SWR[] = {"SWR--.-"};
const char Display_Splt[] = {"Splt"};
const char Display_VFO[] = {"VFO"};
const char Display_JUMA_TRX2[] = {"JUMA-TRX2 "};
const char Display_OH2NLT_OH7SV[] = {"  OH2NLT-OH7SV  "};
const char Display_ChkSumErr[] = {" Checksum Error "};
const char Display_Load_Defaults[] = {"Loading Defaults"};
const char Calibration_Mode[] = {"Calibration Mode"};
const char Display_Next[] = {" DISPLAY:  Next "};
const char Display_Save[] = {"FAST/VFAST: Save"};
const char s9_plus_msg[] = {"S9+%2idB  "};
const char s9_msg[] = {"S%-8i"};
const char Display_Attenuator[] = {"Attenuator:%3idB"};
const char Display_RX_Data[] = {"RX Data: %c  0x%2.2X"};
const char Display_VFO_Lock[] = {" VFO%c%c"};
const char baud_rate_msg[] = {"\n\rUART1        : %i00 Baud\n\r"};
const char Display_System_Clk[] = {"System Clock :%5i kHz\n\r"};
const char VM_Option_Found[] = {"Voice Memory : %sInstalled\n\r"};
const char Display_EEPROM_Chksum[] = {"Cal Checksum :%5.4X\n\rUser Checksum:%5.4X\n\rReset Counter: %d\n\r"};
const char Display_UART2_msg[] = {"JUMA-TRX2 UART2 fixed at 9600 bd\n\r"};
const char eeprom_msg[] = {"\n\rWords   :%5d\n\rChecksum:%5.4X\n\r"};
const char _16s[] = {"%-16s"};
const char Configuration_Msg[] = {" Configuration"};
const char rbs_fmt[] = {"%7s"};

// User Configuration Prompts (Line 2)
const char dp_0_msg[] = {"AGC Speed:%6s"};		// Adjust AGC Speed - Fast/Slow
const char dp_1_msg[] = {"%3s LPF:%5li Hz"};	// Adjust Filter Bandwidths (Narrow/Mid/Wide/Tx)
const char dp_2_msg[] = {"Processor:%6s"};		// Adjust Speech Processor - On/Off
const char dp_3_msg[] = {"Audio I/P:%6s"};		// Select Audio I/P - Mic/Line
const char dp_4_msg[] = {"Keyer:%10s"};			// Select Keyer - Dot Priority/Iambic A/Iambic B/Vibroplex/Straight
const char dp_5_msg[] = {"Delay:%8imS"};		// CW Release Delay (Only in Straight Mode, skipped otherwise)
const char dp_6_msg[] = {"CW Pitch:%4i Hz"};	// Adjust CW Pitch
const char dp_7_msg[] = {"CW Pot:%9s"};			// Select CW Pot Mode, CW Speed/Squelch
const char dp_8_msg1[] = {"Locked:%6iwpm"};		// Default CW Speed Locked
const char dp_8_msg2[] = {"Unlocked:%4iwpm"};	// Default CW Speed Unlocked, adjust with RIT knob
const char dp_9_msg[] = {"CW/CWR Swap:%4s"};	// Adjust CW/CWR Swap On/Off
const char dp_10_msg[] = {"Backlight:%6i"};		// Adjust Backlight
const char dp_11_msg[] = {"Contrast:%7i"};		// Adjust Contrast
const char dp_12_msg1[] = {"%10s%4i S"};		// Adjust LCD Timer 1 - 3600 seconds
const char dp_12_msg2[] = {"%10s%6s"};			// Adjust LCD Timer Off
const char dp_13_msg[] = {"RS232:%10s"};		// Select RS232 Mode - Yaseu CAT/Kenwood Cat/Juma TRX2/Test Mode/Voice Memory
const char dp_14_msg[] = {"Baud Rate:%4i00"};	// Select Baud Rate 1200 - 115200 Baud
const char dp_15_msg1[] = {"%11s(CAT)"};		// Display locked VFO mode, A/B + Split (CAT)
const char dp_15_msg2[] = {"%5s%7i Mem"};		// Adjust number of VFOs - 3 - 26 VFOs
const char dp_16_msg[] = {"S-Meter:%8s"};		// Select S-Meter - Graphic/Numeric
const char dp_17_msg[] = {"Hold Time:%6.2f"};	// Adjust S-Meter Hold Time
const char dp_18_msg[] = {"TX Disable:%5s"};	// Adjust TX Disable - Off/On
const char dp_19_msg[] = {"Auto USB/LSB:%3s"};	// Adjust Auto USB/LSB - Off/On
const char dp_20_msg[] = {"Display:%8s"};		// Select Frequency Display - Original/New/Fixed-A/Fixed-B
const char dp_22_msg[] = {"Include RIT:%4s"};	// Adjust Include RIT On/Off
const char dp_23_msg[] = {"Tune Atten:%3idB"};	// Select Tune Attenuation - 0dB/1dB/2dB/3dB
const char dp_24_msg[] = {"VFO Adj:%8s"};		// Select VFO Adj Off/10Hz/100Hz-10Hz/All
const char dp_5s_11s_fmt[] = {"%5s%11s"};		// Format descriptor used by display pages 15-21-25-26-27

// Common Prompts
const char _mode[] = {"Mode:"};
const char _type[] = {"Type:"};
const char _page[] = {"Page:"};
const char _lcd_timer[] = {"LCD Timer:"};
const char _a_b_split[] = {"A/B + Split"};

// Serial Messages
const char Settings_Saved_EEPROM[] = {"\n\rUser settings saved to EEPROM, Power OFF\n\r"};
const char RS232_Test_Exit[] = {"\n\rRS-232 Test\n\rPress PWR button to exit...\n\r"};
const char Test_Terminated[] = {"\n\rRS-232 Test Terminated\n\r"};
const char Relative_Power[] = {"Rel Power:%6.2fdB "};
const char Press_Hold[] = {"Press and hold the PTT switch to start..."};
const char Equalising[] = {"\n\rEqualising...\n\r"};
const char Averaged_Pwr[] = {"Band:%4im Averaged Power:%5.1f Watts\n\r"};
const char New_Line[] = {"\n\r"};
const char Band_Atten[] = {"Band:%4im Attenuator:%3idB\n\r"};
const char Release_PTT[] = {"Release the PTT switch to complete.\n\r"};
const char Equalisation_Complete[] = {"Equalisation Complete\n\r"};
const char Equalisation_Aborted[] = {"Equalisation Aborted - No changes made.\n\r"};

// Tuning steps: 0 = S (1Hz), 1 = M (10Hz), 2 = F (100Hz), 3 = V (1kHz), and 4 = X (10kHz) (Not Used)
// Avago Technologies HRPG-ASC-11R Optical Rotary Encoder - 120 counts/revolution.
const char tuning_step[] = {			// Indexed by: eeprom.defval.step_idx
						'S',			// 0 - Slow, 1Hz (480Hz/turn)
						'M',			// 1 - Medium, (Default) 10Hz (4.8kHz/turn)
						'F',			// 2 - Fast, 100Hz (48kHz/turn)
						'V',			// 3 - Very Fast, 1kHz (480kHz/turn)
						'X'				// 4 - eXtra Fast, 10kHz (4.8MHz/turn)
						};
// Prompt Messages
const char *s_meter_type[] = {			// Indexed by: eeprom.defval.S_Meter_Type
						"Graphic",
						"Numeric"
						};

const char *agc_speed[] = {				// Indexed by: eeprom.defval.agc_speed
						"Slow",
						"Fast"
						};

const char *audio_input[] = {			// Indexed by: eeprom.defval.af_input
						"Line",
						"Mic"
						};

const char *cw_spd_pot[] = {			// Indexed by: eeprom.defval.sq
						"CW Speed",
						"Squelch-A",
						"Squelch-B"
						};

const char *band_limits[] = {			// Indexed by: cal.calval.Band_Limits
						" IARU Region 1  ",
						"  USA  (Extra)  ",
						" USA  (General) "
						};

const char *on_off[] = {
						"Off",
						"On"
						};

const char *mode_msg[] = {				// Indexed by: eeprom.defval.mode[vfo_idx], eeprom.defval.mode[eeprom.defval.vfo_sel], eeprom.defval.mode[src_vfo]
						"LSB ",
						"USB ",
						"CW  ",
						"CWR ",
						"Tune"
						};

const char *band_sw_msg[] = {			// Do not make these variable length strings. lcd_putst() uses them.
						"Default",		// Indexed by: eeprom.defval.Band_SW
						"   User",
						" Locked"
						};

const char *filter_width[] = {			// Indexed by: eeprom.defval.filter
						"NAR",
						"MID",
						"WID",
						" Tx"
						};

const char *keyer_msg[] = {				// Indexed by: eeprom.defval.keyer
						"Dot Pri.",
						"Iambic A",
						"Iambic B",
						"Vibroplex",
						"Straight"
						};

const char *rs232_mode[] = {			// Indexed by: eeprom.defval.serial_mode
						"Yaesu CAT",
						"KenwoodCAT",
						"JUMA TRX2",
						"Test Mode",
						"Voice Mem"
						};

const char *freq_disp[] = {				// Indexed by: eeprom.defval.Freq_Display
						"Original",
						"New",
						"Fixed-A",
						"Fixed-B"
						};

const char *vfo_adj[] = {				// Indexed by: eeprom.defval.VFO_Adjust
						"Off",
						"10Hz",
						"100/10Hz",
						"All"
						};

const char *power_msg[] = {				// Indexed by: eeprom.defval.nb (power_down_mode)
						"Immediate",
						"Prompted",
						"Save State"
						};

const char *save_settings_prompt[] = {	// Indexed by: flag in function: save_settings(), and in _VM_Command() in serial_ext_kb.c
						" Save Settings? ",
						" Reset Defaults?",
						"Erase Voice Mem?"
						};

const char *disp_type[] =	{			// Indexed by: eeprom.defval.Display_Type
						"Original",
						"Large",
						"Small"
						};

const char *saved_restored[] = {		// Indexed by: local flag in function: save_settings(), as well as _VM_Command() in serial_ext_kb.c
						"   Cancelled!   ",
						"     Saved!     ",
						"     Erased!    "
						};

const char lock_symbol[] = {			// Indexed by: vfo_lock
						' ',			// Unlocked
						'*'				// Locked
						};

const char *cw_rls_speed[] = {			// Indexed by: eeprom.defval.sq
						"CW Speed",
						"Preset",		// Duplicating the entry rather than calculating an index saves 15 bytes! A.Ryan - 5B4AIY - 25/MAY/13
						"Preset"
						};

const char *Not_Installed[] = {
						"Not ",
						""
						};

const char *start_page_select[] = {		// Indexed by masked value of display_page
						"Mode/BW",		// 0
						"O/P Power",	// 1
						"SWR",			// 2
						"Voltage",		// 3
						"PA Current",	// 4
						"Last Page"		// 5
						};

const int filter_beep[]	= {				// Used for filter selection and tuning rate adjustments.
						HZ466_85,		// Indexed by: eeprom.defval.step_idx, eeprom.defval.filter 
						HZ587_31,
						HZ698_45,
						HZ800
						};

/*
Initially, the baud rate text was declared as a series of text strings, which used 6 bytes for the characters
and an additional byte for the string terminator. By declaring the array as a character array and saving the
1200 baud multiplier value (1, 2, 4, 8, 16, 32, 48, and 96), the array size could be reduced from 56 bytes to only
8. To display the value, since either sprintf() or printf() was being used, the formatting descriptor was:
%6li, and the actual value was then: 1200L * eeprom.defval.br[br_txt]. Unfortunately the multiplication added
more code than the saving in memory! By saving instead the integer values of the speed, and then simply adding
the two trailing zeros with a format descriptor of %4i00 the memory savings were considerably improved.
A. Ryan 5B4AIY 11/DEC/2011
*/
const int br_txt[] = {					// Indexed by: eeprom.defval.br
						12,
						24,
						48,
						96,
						192,
						384,
						576,
						1152
					};

/*
 Band prompt digits used in the Rapid Band Switch feature, the first is blank and is never used, but provides
 a place-holder, as the index of valid frequencies is from 1 - 9. If the function returns 0, this is used to
 indicate an out-of-band frequency.
 */
const unsigned char band[] = {
							0,		// 0
							160,	// 1
							80,		// 2
							40,		// 3
							30,		// 4
							20,		// 5
							17,		// 6
							15,		// 7
							12,		// 8
							10		// 9
							};

const unsigned char prompt_index[] = {	// Indexed by page, local variable in display_settings()
							0,		// 0  - "Adjust"
							2,		// 1  - "Adjust Filter BW"
							11,		// 2  - "Speech"
							1,		// 3  - "Select"
							1,		// 4  - "Select"
							5,		// 5  - "CW Release"
							0,		// 6  - "Adjust"
							3,		// 7  - "Select Mode"
							4,		// 8  - "Default CW Speed"
							0,		// 9  - "Adjust"
							0,		// 10 - "Adjust"
							0,		// 11 - "Adjust"
							0,		// 12 - "Adjust"
							1,		// 13 - "Select"
							1,		// 14 - "Select"
							12,		// 15 - "Select VFO"
							1,		// 16 - "Select"
							6,		// 17 - "Adjust S-Meter"
							0,		// 18 - "Adjust"
							0,		// 19 - "Adjust"
							7,		// 20 - "Select Frequency"
							13,		// 21 - "Fast Band Switch"
							0,		// 22 - "Adjust"
							1,		// 23 - "Select"
							1,		// 24 - "Select"
							8,		// 25 - "Meter Scale"
							9,		// 26 - "Power Down"
							10		// 27 - "Start-Up Display"
							};

// User Configuration Prompts (Line 1)
const char *user_cfg_prompts[] = {		// Indexed by: prompt_index[page], local variable in function: display_settings()
							"Adjust",			// 0
							"Select",			// 1
							"Adjust Filter BW",	// 2
							"Select Mode",		// 3
							"Default CW Speed",	// 4
							"CW Release",		// 5
							"Adjust S-Meter",	// 6
							"Select Frequency",	// 7
							"Meter Scale",		// 8
							"Power Down",		// 9
							"Start-Up Display",	// 10
							"Speech",			// 11
							"Select VFO",		// 12
							"Fast Band-Switch"	// 13
							};

// Amateur Band Limits for Fast Band Switch and TX Out-Of-Band Inhibit
long upper_limit[] = {
						_160M_UPPER,		// 0 - Place holder
						_160M_UPPER,		// 1
						_80M_UPPER_IARU,	// 2
						_40M_UPPER_IARU,	// 3
						_30M_UPPER,			// 4
						_20M_UPPER,			// 5
						_17M_UPPER,			// 6
						_15M_UPPER,			// 7
						_12M_UPPER,			// 8
						_10M_UPPER			// 9
						};

 long lower_limit[] = {
						_160M_LOWER,		// 0 - Place holder
						_160M_LOWER,		// 1
						_80M_LOWER,			// 2
						_40M_LOWER,			// 3
						_30M_LOWER,			// 4
						_20M_LOWER,			// 5
						_17M_LOWER,			// 6
						_15M_LOWER,			// 7
						_12M_LOWER,			// 8
						_10M_LOWER			// 9
						};

// Centre frequencies of the amateur bands from 160m - 10m. Used by the Fast bandswitch feature.
const unsigned long amateur_bands[] = {
									_160M_CENTRE,	// 0
									_80M_CENTRE,	// 1
									_40M_CENTRE,	// 2
									_30M_CENTRE,	// 3
									_20M_CENTRE,	// 4
									_17M_CENTRE,	// 5
									_15M_CENTRE,	// 6
									_12M_CENTRE,	// 7
									_10M_CENTRE		// 8
									};

// Frequency select points for Fast Bandswitch feature.
const unsigned long band_select_freq[] = {
									0L,				// 0 - Dummy value, never used.
									SELECT_160,		// 1
									SELECT_80,		// 2
									SELECT_40,		// 3
									SELECT_30,		// 4
									SELECT_20,		// 5
									SELECT_17,		// 6
									SELECT_15,		// 7
									SELECT_12		// 8
									};

// US General Class lower segment limit
const unsigned long lower_segment_limit[] = {
									30000000UL,		// 0 - Dummy value, place holder
									30000000UL,		// 1 - 160m Dummy value, place holder
									3600000UL,		// 2 - 80m
									7125000UL,		// 3 - 40m
									3000000UL,		// 4 - 30m Dummy value, place holder
									14150000UL,		// 5 - 20m
									30000000UL,		// 6 - 17m Dummy value, place holder
									21200000UL,		// 7 - 15m
									30000000UL,		// 8 - 12m
									30000000UL		// 9 - 10m
									};

// US General Class upper segment limits
const unsigned long upper_segment_limit[] = {
									0UL,			// 0 - Dummy value, place holder
									0UL,			// 1 - 160m Dummy value, place holder
									3800000UL,		// 2 - 80m
									7175000UL,		// 3 - 40m
									0UL,			// 4 - 30m Dummy value, place holder
									14225000UL,		// 5 - 20m
									0UL,			// 6 - 17m Dummy value, place holder
									21275000UL,		// 7 - 15m
									0UL,			// 8 - 12m
									0UL				// 9 - 10m
									};

// SCAF filter tuning limits
const int scaf_min[] = {			// Upper Limit. Indexed by: eeprom.defval.filter
						NAR_SCAF,	// 0 - Narrow	1,300Hz
						MID_SCAF,	// 1 - Mid		2,300Hz
						MIN_SCAF,	// 2 - Wide		4,000Hz
						TX_UPPER	// 3 - Transmit	3,750Hz
						};

const int scaf_max[] = {			// Lower Limit. Indexed by: eeprom.defval.filter
						MAX_SCAF,	// 0 - Narrow	500Hz
						NAR_SCAF,	// 1 - Mid		1,300Hz
						MID_SCAF,	// 2 - Wide		2,300Hz
						TX_LOWER	// 3 - Transmit	2,272Hz
						};
// ================================================================================================================
void (*RS_232_Mode)(void);		// RS-232 Mode, Yaesu CAT/Kenwood CAT/TRX-2/Voice Memory/Test
const char *user_band_prompt;	// Pointer to the characters for the User Band prompt messages

// Main Board 74HC595 I/O latch
unsigned int hc595 = 0;			// Latch bits defined in juma-trx2.h file

/*
  Global DDS Variables
  The freq_in_use variable always holds the DDS frequency. In receive it will hold the actual carrier frequency
  for USB and LSB. In the CW modes, (CW / CWR), the DDS frequency is offset by the amount of the CW shift. If
  the mode is CW, the transceiver is set to USB, and the local oscillator is offset lower in frequency by the
  CW shift. Conversely, if the mode is CWR the transceiver is set to LSB, and the local oscillator is offset
  higher in frequency by the amount of the CW shift. If RIT is turned on, then the frequency is further offset
  by the RIT shift.
  In transmit, it will hold the carrier frequency. It is also the frequency that is displayed in both transmit
  and receive. In the receive mode, if Include RIT is on, then this variable is displayed, otherwise it is the
  VFO frequency.
*/
long freq_in_use;				// Frequency currently in use, always used to load the DDS
int vfo_idx;					// VFO index, A = 0, B = 1, if (Split) then  RX = 0, TX = 1
int rit;						// RIT offset frequency (+/- 1kHz maximum.)
int tmp_sub_page0 = 0;			// Used to restore the subpage when blinking the mode change
int mode_ctr = 0;				// Used as the Mode/Filter blink counter
int rf_filt_idx;				// Index to filter & attenuator bits table
int old_ADC = 2048;				// Used in get_rit()

// DDS & Clock Variables
int fcy;						// Clock / kHz
long tx_filter_bw;				// Transmit filter bandwidth upper limit

// System Status Flags
int tx_enable = ENABLE;			// 1 = Enable, 0 = Disable
int cw_active = OFF;			// Status flag for keyer module (timers_pwm.c)
int tune_on = OFF;				// Tune state indicator
// Preset transceiver to RX mode. (Prevents initial spurious TX on switch-on.) A.Ryan - 5B4AIY - 21/AUG/2013
int PTT_IN = RECEIVE;			// Copy of PTT_IN_IO input / TX ON, 0 = Active
int dsp1_ctr = 0;				// Used for display blink feature
int dsp2_ctr = 0;
int band_in_use = NOT_HAM_BAND;
int split_flag = FALSE;
int svc_flag = FALSE;			// Used to inhibit the drain current hold timer in the Service mode
int latch = FALSE;

/*
  Alarm System Flag
  0 = No Alarm
  0x01 = Overcurrent
  0x02 = SWR
  0x04 = Overvoltage
  0x08 = Undervoltage
  0x10 = Undervoltage Pre-Limit
  0x2x Display Blank flag set
*/
int alarm_flag = FALSE;			// Alarm Flag, 0 = No Alarms
int loop_ctr = 0;				// Alarm Display Blink Counter 0 = Reset
int adj_attn = FALSE;			// Set when adjusting attenuator
//int loop_ctr2 = 0;				// Debug Main Loop timing counter - comment out when not needed - 5B4AIY

// Timing
int seconds = 0;				// Free running 1-sec timer, used by ISD1700 routines

// Keyer
int keyer_mode = -1;			// 0 = Dot priority, 1 = Iambic-A, 2 Iambic-B, 3 = Straight, -1 = Startup inhibit (Jan Wieck WI1CK)

// Voice memory option
int isd_1700_installed = FALSE;	// Voice memory board 0 = Not Installed

// System Status Flags
char lcdpbuff[20];				// LCD print buffer, longest display is 16 characters, this is a safety margin.
int	lcd_mode = NORMAL_DISPLAY;	// LCD display mode
int sub_page0 = 0;				// LCD display sub page for lcd_mode0
int sub_page1 = 0;				// LCD display sub page for lcd_mode1 - AGC Speed
int vfo_lock = OFF;				// VFO knob tune lock
int mute = AUDIO_ON;			// Mute test flag, 1 = Mute, 0 = Audio on
int changed_flag = FALSE;		// Set if User Memory or VFO Memory is changed, cleared in save_defaults().

// ADC & Meter Variables
double batt_mean_val;			// Battery Voltage x 1,000,000
int batt_raw;					// Raw A-D convertor output
int batt_scale_factor;			// Bar Graph Battery Voltage scale factor
int s_meter;					// Scaled S-meter value 0 ... 48
int s_meter_timer;				// Hold time for numeric S-Meter
int s_meter_const;				// S-Meter constant, set in main()
int fwd_adval;					// Forward & Reverse power raw values, measured only when in TX mode
int rev_adval;
long fwd_pwr;					// Relative forward power
long rev_pwr;					// Relative reverse power
long out_pwr;					// Scaled output power = (fwd_pwr - rev_pwr) * scale
double power_bar_graph_const;	// Used in display_power_meter()
int id_raw_value;				// Drain current A-D value
int id_const;					// Bar Graph Drain Current scaling constant
int swr;						// Calculated SWR * 100
int swr_al_dly;					// SWR alarm activation delay counter, main loop cycles.
int pwr_meter;					// Scaled output power meter bar-graph value 0 ... 48, used in TRX-2 protocol

int cw_pot;						// CW speed potentiometer, also used for squelch
int sq_timer;					// Squelch open timer
int squelch_on;					// Squelch on flag, 0 = OFF, 1 = ON

int bl_timer;					// LCD back light timer(s), 0 = Not In Use
int bl_off_flag;				// Turn backlight off flag, 1 = BL OFF

int glitch_time;				// Glitch removal timer value. For MODE switch changes, 5mS, for VFO changes 50mS
int low_voltage_warning = OFF;	// Pre-shutoff warning voltage, raw ADC value, = Low Voltage Alarm + 100mV

// User power on values, EEPROM storage 
int fd_counter;					// Counts how many times factory defaults have reloaded - Changed to int - 5B4AIY
// ================================================================================================================
// User Configuration Settings
union
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval) / 2];
	} eeprom;

// System Calibration Values
union
	{
	struct calval calval;
	unsigned int ee[sizeof(struct calval) / 2];
	} cal;
// ================================================================================================================
// Save User Configuration Data
void save_defaults(void)
	{
	int i;

	if((display_page & 0x07) == LAST_PAGE && lcd_mode == NORMAL_DISPLAY)
		display_page = LAST_PAGE + (sub_page0 << 4);	// Save the current display page

	eeprom.defval.d_csum = 0;

	for(i = 0; i < (sizeof(struct defval) / 2); i++)	// Size is returned in bytes, (316) but array eeprom.storage[] is in words.
		{
		if(i < ((sizeof(struct defval) / 2) - 1))
			eeprom.defval.d_csum += eeprom.storage[i];							// Calculate mod 0xFFFF checksum
		else
			eeprom.defval.d_csum = -eeprom.defval.d_csum;						// Last round, invert checksum

 		EraseEE(EEPAGE, ((2 * i) + EEDEF), WORD);
		WriteEE((int *)&(eeprom.storage[i]), EEPAGE, ((2 * i) + EEDEF), WORD);	// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits - Added cast 5B4AIY
		}

	if(eeprom.defval.serial_mode == SERIAL_TEST)
		printf(eeprom_msg, i, eeprom.defval.d_csum);	// Only print in RS-232 Test Mode

	changed_flag = FALSE;
	}

// Read User Configuration Data
unsigned int read_defaults(void)
	{
	int i;
	unsigned int checksum = 0;

	for(i = 0; i < (sizeof(struct defval) / 2); i++)	// Size is returned in bytes, (316) but array eeprom.storage[] is in words.
		{
		ReadEE(EEPAGE, ((2 * i) + EEDEF), (int *)&(eeprom.storage[i]), WORD);	// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits - Added cast 5B4AIY
		checksum += eeprom.storage[i];
		}

// Read factory defaults reset counter
	ReadEE(EEPAGE, EE_FD_LOC, &fd_counter, WORD);

	return checksum;
	}

// Save Calibration Values
void save_calval(int prt_flag)
	{
	int i;

	cal.calval.c_csum = 0;

	for(i = 0; i < (sizeof(struct calval) / 2); i++)							// Size is returned in bytes, (46) but array cal.ee[] is in words.
		{
		if(i < ((sizeof(struct calval) / 2) - 1))
			cal.calval.c_csum += cal.ee[i];									 	// Calculate mod 0xFFFF checksum
		else
			cal.calval.c_csum = -cal.calval.c_csum;								// Last round, invert checksum

		EraseEE(EEPAGE, ((2 * i) + EECAL), WORD);
		WriteEE((int *)&(cal.ee[i]), EEPAGE, ((2 * i) + EECAL), WORD);			// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits - Added cast 5B4AIY
		}

	if(prt_flag && (eeprom.defval.serial_mode == SERIAL_TEST))
		printf(eeprom_msg, i, cal.calval.c_csum);	// Only print in Test Mode
	}

// Read Calibration Values
unsigned int read_calval(void)
	{
	int i;
	unsigned int checksum;

	checksum = 0;

	for(i = 0; i < (sizeof(struct calval) / 2); i++)					// Size is returned in bytes, (46) but array cal.ee[] is in words.
		{
		ReadEE(EEPAGE, ((2 * i) + EECAL), (int *)&(cal.ee[i]), WORD);	// EEPROM Address 8 high bits, address + physical EEPROM start 16 low bits - Added cast 5B4AIY
		checksum += cal.ee[i];
		}

	return checksum;
	}

// Set factory defaults for EEPROM storage. (See trx2_eeprom.h for mapping details.)
void set_factory_defaults(void)
	{
	int i, m;
// User settings
	eeprom.defval.vfo[VFO_A] = DEFAULT_FREQ_A - 5000;	// 3.505MHz
	eeprom.defval.vfo[VFO_B] = DEFAULT_FREQ_A;			// 3.6MHz
	eeprom.defval.mode[VFO_A] = LSB;					// Lower Sideband
	eeprom.defval.mode[VFO_B] = LSB;					// Lower Sideband

// Fill the remaining VFO memories with the amateur band frequencies
	for(i = 2, m = 0; i < 30; i++, m++)
		{
		eeprom.defval.vfo[i] = amateur_bands[i % 9];
		m %= 9;

		if(m < 3)
			eeprom.defval.mode[i] = LSB;			// For 160 - 40, Lower Sideband
		else
			eeprom.defval.mode[i] = USB;			// For 30 - 10, Upper Sideband
		}

// Fill the remaining frequencies, to be used as user band frequencies, with the amateur band frequencies
	for(i = 30, m = 0; i < 39; i++, m++)
		{
		eeprom.defval.vfo[i] = amateur_bands[m];	// Fill other memory locations with amateur band centre frequencies

		if(m < 3)
			eeprom.defval.mode[i] = LSB;			// For 160 - 40, Lower Sideband
		else
			eeprom.defval.mode[i] = USB;			// For 30 - 10, Upper Sideband
		}

	eeprom.defval.vfo[39] = 0;						// Spare Location (Not Used - 4 bytes)
	eeprom.defval.mode[39] = MODE_BW;				// Spare Location. Now used as the start-up display page. Default = Mode/BW
	eeprom.defval.step_idx = TEN_HZ;				// 10Hz tuning rate
	eeprom.defval.vfo_sel = VFO_A;					// VFO-A
	eeprom.defval.max_vfo = SPLIT_MODE;				// A/B + Split
	eeprom.defval.rit_on = OFF;						// RIT OFF
	eeprom.defval.agc_speed = SLOW;					// AGC Slow
	eeprom.defval.proc_on_off = OFF;				// Speech Processor OFF
	eeprom.defval.nb = PROMPTED;					// Datapoint now used for Fast Power Off feature. 0 = Immediate, 1 = Prompted, 2  = Save State
	eeprom.defval.filter = WIDE;					// WID, for Matti
	eeprom.defval.contrast = DEFAULT_CONTRAST;		// 2000
	eeprom.defval.back_light = DEFAULT_BL;			// 350
	eeprom.defval.back_light_timer = OFF;
	eeprom.defval.scaf_filters[NARROW] = DEFAULT_NAR;
	eeprom.defval.scaf_filters[MEDIUM] = DEFAULT_MID;
	eeprom.defval.scaf_filters[WIDE] = DEFAULT_WID;
	eeprom.defval.scaf_filters[TX_FILTER] = DEFAULT_TX;
	eeprom.defval.cw_shift = CW_SHIFT;				// 700Hz
	eeprom.defval.cw_sidetone = CW_SIDETONE;		// 700Hz
	eeprom.defval.keyer = IAMBIC_B;					// Iambic B
	eeprom.defval.serial_mode = SERIAL_TEST;		// Test Mode (3)
	eeprom.defval.af_input = MIC;
	eeprom.defval.sq = OFF;							// Squelch not in use
	eeprom.defval.br = DEFAULT_BAUD_RATE;			// Initial Baud rate = 9600
	eeprom.defval.S_Meter_Type = GRAPHIC;			// Graphic S-Meter - 5B4AIY
	eeprom.defval.OOB_Check = ENABLE;				// Out-Of-Band Check flag 0 = Disable, 1 = Enable
	eeprom.defval.Auto_SB = ENABLE;					// Auto Sideband Select 0 = Disable, 1 = Enable
	eeprom.defval.Freq_Display = FIXED_B;			// Frequency Display: 0 = Old, 1 = New, 2 = Fixed 1Hz, but no leading zeros, and decimal point in either Mhz or kHz, 3 = Fixed 10Hz, no leading zeros
	eeprom.defval.Band_SW = _DEFAULT;				// Band Switch Mode 0 = Default, 1 = User, 2 = Locked
	eeprom.defval.RIT_Include = NO;					// Include RIT shift in frequency display, 0 = NO, 1 = YES
	eeprom.defval.Tune_Attenuation = MAX_ATTEN;		// Set attenuation in TUNE mode to 3dB
	eeprom.defval.VFO_Adjust = OFF;					// Adjust VFO to nearest decade step. 0 = OFF, 1 = 10Hz, 2 = 10/100Hz, 3 = All
	eeprom.defval.Hold_Time = 2500;					// Numeric S-Meter peak hold time (Approx 2.5 seconds)
	eeprom.defval.CW_Speed = (2804 | LOCKED);		// 20wpm and Locked
	eeprom.defval.VFO_Lock = 0;						// All VFOs unlocked
	eeprom.defval.CW_CWR_Swap = OFF;				// Original operation, FILTER button does not swap CWR back to CW
	eeprom.defval.Display_Type = ORIGINAL_MARKERS;	// Graphic Scale Type 0 = Original, 1 = Large Markers, 2 Small Markers
	eeprom.defval.CW_Release = DEFAULT_CW_RELEASE;	// Default CW Release time, 500mS
/*
  The VFO_Lock variable is an unsigned long integer. Each bit position holds the Lock/Unlock status of its
  corresponding VFO memory. In other words, bit 0 holds the lock status of VFO-A, bit 1, VFO-B etc. This
  allows for selective locking of individual VFO memories. It was tempting to use a bit in the mode array,
  but as the variable eeprom.defval.mode[] is referenced 23 times in the code, it would have meant a considerable
  amount of extra code to mask off the lock bit when it was not required, as well as OR and AND it when it was.
  Using this dedicated integer is far simpler, and less computationally intensive. - Adrian Ryan 20/NOV/2011
*/
// Calibration values
	cal.calval.ref_osc = REF_OSC;
	cal.calval.id_mult = ID_MULT;
	cal.calval.batt_mult = BATT_MULT;
	cal.calval.s_mtr_cal = S_METER_CAL;
	cal.calval.fwd_pwr_mult = FWD_PWR_MULT;
	cal.calval.beep_len = DEFAULT_BEEP_LEN;
	cal.calval.rf_atn_bits[0] = MAX_ATTEN;			// 250 kHz  Initial values for RF attenuator bits
	cal.calval.rf_atn_bits[1] = MAX_ATTEN;			// 500 kHz  Valid range 0 ... 3
// The size of array cal.calval.rf_atn_bits[] is 16 bytes. Entries 11 - 15 are spare, currently unused.
	for(i = 2; i < 12; i++)
		cal.calval.rf_atn_bits[i] = 1;

	cal.calval.swr_trip = SWR_ALARM;				// SWR = 3.00
	cal.calval.swr_delay = LONG_SWR_DLY;			// Default Long SWR Delay Time, 5 Seconds
	cal.calval.alarm_flags = 0x0F;					// Enable all alarms
	cal.calval.overvoltage_trip = OVERVOLTAGE_TRIP;	// Over-voltage trip (Nominally 2762 = 14.5V)
	cal.calval.overcurrent_trip = OVERCURRENT_TRIP;	// Over-current trip (Nominally 2463 = 2.5A)
	cal.calval.undervoltage_trip = UNDERVOLTAGE_TRIP;	// Under-voltage trip (Nominally 2095 = 11.0V)
	cal.calval.Band_Limits = IARU_1;				// 0 = IARU Region 1, 1 = USA Extra Class, 2 = USA General Class

// Increment factory default reset counter
	ReadEE(EEPAGE, EE_FD_LOC, &fd_counter, WORD);
	fd_counter++;

// Save factory defaults reset counter
	EraseEE(EEPAGE, EE_FD_LOC, WORD);
	WriteEE(&fd_counter, EEPAGE, EE_FD_LOC, WORD);
	save_calval(1);					// Save Calibration Settings
	save_defaults();				// Save Configuration settings
	}

void display_screen(const char *line_1, const char *line_2)
	{
	clear_lcd();
	lcd_putst(line_1);
	lcd_cmd(LINE2);
	lcd_putst(line_2);
	}

int set_mode_blink(void)
	{
	tmp_sub_page0 = sub_page0;		// Save the current page for later restoral.
	sub_page0 = 0;					// Set the sub page to zero to display the mode and filter settings
	mode_ctr = MODE_COUNTER;		// Reset the mode counter so that we get two blinks,
	return 3;						// and return the value for the dsp_ctr variable.
	}
/*
  Separate CW Speed & Squelch feature
  By checking whether the pot value has changed since the last pass through the main loop, it is now possible
  to avoid having the CW Speed changed when altering the squelch level. When the CW Speed pot function is set
  to Squelch, the CW Speed defaults to the user settable preset speed. When the function is set to CW Speed,
  the pot directly controls the speed, except if it is rotated fully anti-clockwise, whereupon the default preset
  speed is selected. To ensure that all pots do this, there is a 10-count dead band. If the A-D convertor O/P
  is less than 10, then it is in the dead band, and the default speed selected. Adrian Ryan 20/NOV/2011
  The speed, in WPM = 1200 / Time in mSec
  To set the CW speed to the default value, note the minimum pot value is 10. Use these constants:

  Speed Range,wpm		Division Constant		Addition Constant		Default
  ================================================================================
	10 - 30					208692					40					2053/20wpm
	 5 - 40					79501					30					2551/20wpm
	 5 - 30					83477					40					2804/20wpm
	 6 - 60					92752					20					2169/20wpm
	10 - 60					166953					20					1511/20wpm
  ================================================================================
*/
int cw_speed(int speed)
	{
	long time;

	time = (speed < 10) ? (long)(4096 - (eeprom.defval.CW_Speed & UNLOCKED)) : (long)(4096 - speed);
	time *= time;								// Make action non-linear, Min = 1, Max = 16,695,396 (Pot = 10)
	time /= 83477L;								// Division Constant
	time += 40;									// Addition Constant - Sets maximum speed
	return (int)time;
	}

int select_nearest_band(long freq)				// Used in the Rapid Bandswitch routine to select the nearest band
	{											// Valid bands are 1 - 9
	int i = 1;

	do	{
		if(freq < band_select_freq[i])
			break;

		i++;
		} while(i < HIGHEST_BAND);

	return i;
	}

void set_tx_filter_bw(void)
	{
	tx_filter_bw = (long)(SCAF_CLK / (long)eeprom.defval.scaf_filters[TX_FILTER]);
	}

// User Configuration Display Pages
void dp_0(void)		// 0 - AGC Speed (Fast/Slow)
	{
	sprintf(lcdpbuff, dp_0_msg, agc_speed[eeprom.defval.agc_speed]);
	}

void dp_1(void)		// 1 - Narrow/Medium/Wide/Tx Filter LPF Bandwidth
	{
	sprintf((char *)lcdpbuff, dp_1_msg, filter_width[eeprom.defval.filter], (long)(SCAF_CLK / (long)eeprom.defval.scaf_filters[eeprom.defval.filter]));	// Calculate & display corner frequency
	}

void dp_2(void)		// 2 - Speech Processor (On/Off)
	{
	sprintf(lcdpbuff, dp_2_msg, on_off[eeprom.defval.proc_on_off]);
	}

void dp_3(void)		// 3 - Audio I/P (Microphone/Line)
	{
	sprintf(lcdpbuff, dp_3_msg, audio_input[eeprom.defval.af_input]);
	}

void dp_4(void)		// 4 - Key Type (Dot Priority/Iambic A/Iambic B/Vibroplex/Straight)
	{
	sprintf(lcdpbuff, dp_4_msg, keyer_msg[eeprom.defval.keyer]);
	}

void dp_5(void)		// 5 - CW Release Time
	{
	if(eeprom.defval.CW_Release > MIN_CW_RELEASE)					// If release time is set,
		sprintf(lcdpbuff, dp_5_msg, eeprom.defval.CW_Release);		// display set release time,
	else
		sprintf(lcdpbuff, CW_rls_prompt, cw_rls_speed[eeprom.defval.sq]);
	}

void dp_6(void)		// 6 - CW Pitch (300Hz - 1100Hz)
	{
	sprintf((char *)lcdpbuff, dp_6_msg, eeprom.defval.cw_shift);	// Display CW shift (Hz)
	}

void dp_7(void)		// 7 - CW Pot Mode (CW Speed/Squelch-A/Squelch-B)
	{
	sprintf(lcdpbuff, dp_7_msg, cw_spd_pot[eeprom.defval.sq]);
	}

void dp_8(void)		// 8 - Default CW Speed Setting
	{
	if(eeprom.defval.CW_Speed & LOCKED)
		sprintf(lcdpbuff, dp_8_msg1, 1200 / cw_speed(eeprom.defval.CW_Speed & UNLOCKED));
	else
		{
		eeprom.defval.CW_Speed = (cw_pot < 10) ? 10 : cw_pot;
		cw_period = cw_speed(eeprom.defval.CW_Speed);
		sprintf(lcdpbuff, dp_8_msg2, 1200 / cw_period);
		}
	}

void dp_9(void)		// 9 - CW/CWR swap using FILTER button
	{
	sprintf((char *)lcdpbuff, dp_9_msg, on_off[eeprom.defval.CW_CWR_Swap]);
	}

void dp_10(void)	// 10 - LCD Backlighting
	{
	sprintf((char *)lcdpbuff, dp_10_msg, eeprom.defval.back_light);
	}

void dp_11(void)	// 11 - LCD Constrast
	{
	sprintf((char *)lcdpbuff, dp_11_msg, eeprom.defval.contrast);
	}

void dp_12(void)	// 12 - LCD Backlight Timer
	{
	if(eeprom.defval.back_light_timer)
		sprintf((char *)lcdpbuff, dp_12_msg1, _lcd_timer, eeprom.defval.back_light_timer * 5);
	else
		sprintf((char *)lcdpbuff, dp_12_msg2, _lcd_timer, on_off[0]);
	}

void dp_13(void)	// 13 - Serial Port Mode (Yaesu CAT/Kenwood CAT/TRX-2/Test/Voice Memory)
	{
	sprintf(lcdpbuff, dp_13_msg, rs232_mode[eeprom.defval.serial_mode]);
	}

void dp_14(void)	// 14 - Serial Port Speed (1200 Baud - 115200 Baud)
	{
	sprintf((char *)lcdpbuff, dp_14_msg, br_txt[eeprom.defval.br]);	// Display current selection
	}

void dp_15(void)	// 15 - VFO Mode (CAT/Split/Multi-Memory)
	{
	if(eeprom.defval.serial_mode < 2)								// If either CAT mode is selected
		sprintf(lcdpbuff, dp_15_msg1, _a_b_split);
	else
		{
		if(eeprom.defval.max_vfo == 2)								// A + B / Split
			sprintf((char *)lcdpbuff, dp_5s_11s_fmt, _mode, _a_b_split);
		else
			sprintf((char *)lcdpbuff, dp_15_msg2, _mode, eeprom.defval.max_vfo);	// Other VFOs
		}
	}

void dp_16(void)	// 16 - S-Meter Type (Graphic/Numeric)
	{
	sprintf(lcdpbuff, dp_16_msg, s_meter_type[eeprom.defval.S_Meter_Type]);
	}

void dp_17(void)	// 17 - Numeric S-Meter Hold Time (0.25 Sec - 5 Sec)
	{
	sprintf(lcdpbuff, dp_17_msg, ((double)eeprom.defval.Hold_Time) / 1000.0);
	}

void dp_18(void)	// 18 - Out-Of-Band TX Disable (On/Off)
	{
	sprintf(lcdpbuff, dp_18_msg, on_off[eeprom.defval.OOB_Check]);
	}

void dp_19(void)	// 19 - Auto USB/LSB Select (On/Off)
	{
	sprintf(lcdpbuff, dp_19_msg, on_off[eeprom.defval.Auto_SB]);
	}

void dp_20(void)	// 20 - Frequency Display (Original/New/Fixed-A/Fixed-B)
	{
	sprintf(lcdpbuff, dp_20_msg, freq_disp[eeprom.defval.Freq_Display]);
	}

void dp_21(void)	// 21 - Rapid Bandswitch Mode (Default/User/Locked)
	{
	sprintf(lcdpbuff, dp_5s_11s_fmt, _mode, band_sw_msg[eeprom.defval.Band_SW]);
	}

void dp_22(void)	// 22 - Include RIT (On/Off)
	{
	sprintf(lcdpbuff, dp_22_msg, on_off[eeprom.defval.RIT_Include]);
	}

void dp_23(void)	// 23 - Tune Attenuation (0dB - 3dB)
	{
	sprintf(lcdpbuff, dp_23_msg, -eeprom.defval.Tune_Attenuation);
	}

void dp_24(void)	// 24 - VFO Frequency Rounding (Off/10Hz/10Hz-100Hz/All)
	{
	sprintf(lcdpbuff, dp_24_msg, vfo_adj[eeprom.defval.VFO_Adjust]);
	}

void dp_25(void)	// 25 - Graphic Scale Type
	{
	sprintf(lcdpbuff, dp_5s_11s_fmt, _type, disp_type[eeprom.defval.Display_Type]);
	}

void dp_26(void)	// 26 - Power Off Mode Selection (Immediate/Prompted/Save State)
	{
	sprintf(lcdpbuff, dp_5s_11s_fmt, _mode, power_msg[power_down_mode]);
	}

void dp_27(void)	// 27 - Start-Up Page Select
	{
	sprintf(lcdpbuff, dp_5s_11s_fmt, _page, start_page_select[display_page & 0x07]);
	}

void (*display_pages[])(void) = {
							dp_0,
							dp_1,
							dp_2,
							dp_3,
							dp_4,
							dp_5,
							dp_6,
							dp_7,
							dp_8,
							dp_9,
							dp_10,
							dp_11,
							dp_12,
							dp_13,
							dp_14,
							dp_15,
							dp_16,
							dp_17,
							dp_18,
							dp_19,
							dp_20,
							dp_21,
							dp_22,
							dp_23,
							dp_24,
							dp_25,
							dp_26,
							dp_27
							};
/*
  Save Current Settings with prompt
  The prompt index displays the type of prompt, the mode takes the values: 1 = Save Calibration settings,
  2 = Save Configuration Settings, 4 = Restore & Save complete factory default calibration & configuration settings.
  Note: Values 1 and 2 may be combined to save both the calibration and the configuration settings.
*/
int save_settings(int prompt, int mode)
	{
	int flag = TRUE;

	display_screen(save_settings_prompt[prompt], PWR_No_MODE_Yes);

	while(PWR_SW || _MODE);					// Wait for PWR or MODE button to be released

	ms_delay(BUTTON_DEBOUNCE);				// Added to prevent spurious cancel command

	while(SW5 && !PWR_SW);					// Wait for a MODE or PWR button push
		
	if(_MODE)								// If MODE button was pressed,
		{
		if(mode & 1)
			save_calval(1);					// Save new System Calibration settings,

		if(mode & 2)						// Save new User Configuration Settings.
			save_defaults();

		if(mode & 4)						// Complete factory default restore and save
			set_factory_defaults();

		beep(HZ587_31, cal.calval.beep_len);
		}
	else
		{
		if(mode & 1)
			read_calval();					// Restore previous System Calibration Settings

		if(mode & 2)
			read_defaults();				// Restore previous User Configuration Settings

		beep(HZ466_85, cal.calval.beep_len);
		flag = FALSE;
		}

	set_tx_filter_bw();
	lcd_cmd(LINE2);
	lcd_putst(saved_restored[flag]);
	ms_delay(1000);							// Display message
	set_chgen(eeprom.defval.Display_Type);	// Load bar graph fonts. Do not move - otherwise LCD is cleared!
	return flag;
	}

void max_min(int *value, int max, int min)
	{
	if(*value > max) *value = max;
	if(*value < min) *value = min;
	}

long int_round(long dividend, long  divisor)
	{
	long q, r;

	q = dividend / divisor;
	r = dividend % divisor;

	if(r >= (divisor >> 1))
		q++;

	return q;
	}

// Calculate SWR
int calc_swr(int f, int r)
	{
	long Ro;
	
	if(f < PWR_MTR_DEAD_BAND)
		return 0;				// Only calculate if we have some power
	
	Ro = ((long)r * 100L) / (long)f;
	return (int)(((100 + Ro) * 100) / (100 - Ro));
	}

int check_freq(long upper)
	{
	int i = 1, band = 0;
	long lower;
	
	lower = upper;

	if(eeprom.defval.mode[vfo_idx] == LSB)
		lower -= tx_filter_bw;

	if(eeprom.defval.mode[vfo_idx] == USB)
		upper += tx_filter_bw;

// CW/CWR/Tune modes do not need correcting, as the displayed frequency is the carrier frequency
	do	{
		if(upper <= upper_limit[i] && lower >= lower_limit[i])
			{
			band = i;
			break;
			}

		i++;
		} while(i < 10);

	if(cal.calval.Band_Limits == USA_GENERAL)	// If General Class set, perform extra checks
		{
		if(upper > lower_segment_limit[band] && lower < upper_segment_limit[band])
			band = 0;
		}

	return band;				// If out-of-band, 0, otherwise 1 - 9
	}
/*
  Calculate & display bar graph S-Meter
  For numeric S-Meter, we will need a flag. Power meter or S-meter taken care of in disp_bar_meter()
  The Numeric/Graphic S-Meter flag will have to be set/cleared in the User Setup section, and the status
  saved to the EEPROM. The peak hold feature was added because the original reacted rather too fast to
  enable reliable readings. The peak hold feature allows the hold time to be set from zero to about
  2.5 seconds in 5mSec increments, and results in a much better presentation. Adrian Ryan 16/NOV/2011
*/
char set_scale(void)
	{
	static int peak_s_meter;			// Initialised to zero by default.

	if(!s_meter_timer)					// Hold timer has timed-out,
		peak_s_meter = 0;				// so reset peak S-meter,

	if(s_meter > peak_s_meter)
		{
		peak_s_meter = s_meter;						// Save new peak value
		s_meter_timer = eeprom.defval.Hold_Time;	// Reset hold timer
		}

	if (peak_s_meter > 54) return 40;
	if (peak_s_meter > 44) return 30;
	if (peak_s_meter > 36) return 20;
	if (peak_s_meter > 29) return 10;
	if (peak_s_meter > 24) return 9;
	if (peak_s_meter > 20) return 8;
	if (peak_s_meter > 18) return 7;
	if (peak_s_meter > 15) return 6;
	if (peak_s_meter > 12) return 5;
	if (peak_s_meter > 8) return 4;
	if (peak_s_meter > 3) return 3;

	return peak_s_meter;
	}

void display_s_meter(int type)
	{
	char s;

	s_meter = (int)int_round((long)convert_adc12(S_METER), (long)cal.calval.s_mtr_cal);
	s = set_scale();

	if(type)										// 0 = Graphic, 1 = Numeric
		{											// New numeric S-Meter
		if(s > 9)
			sprintf(lcdpbuff, s9_plus_msg, s);
		else
			sprintf(lcdpbuff, s9_msg, s);

		lcd_putst(lcdpbuff);
		}
	else											// Original Graphic S-Meter
		draw_s_meter(s_meter);
	}

// Calculate & display bar graph power meter. This now indicates true power.
/*
  Power Bar Graph Constant
  The standard output voltage from the forward power detector is 2.05V at 10W. This gives an A-D output of 4096 * 2.05 / 5.0
  which is a count of 1679. This is squared to give a power value, and is 2,819,041. For the digital display, this value is
  multiplied by the calibration constant stored in cal.calval.fwd_pwr_mult, and whose default value is 3550. This then gives
  a value of: 10,004,045,550. This is displayed as 10.0W by using the sprintf() function and dividing the value by 10^9.
  To quantise the raw value to give a full-scale value of 48 thus requires a constant of 10,004,045,550 / 48 = 208,417,615
  Since the power meter calibration constant has been taken into account, the constant is thus: 208,417,615 / 3550 = 57,864
  Since the out_pwr has already been multiplied by the calibration constant, in order to take this into account the actual
  constant is calculated at boot time, and thus becomes: cal.calval.fwd_pwr_mult / (57,864 * cal.calval.fwd_pwr_mult)
  which is: 205,417,200. The bar graph power meter is thus automatically scaled for 10W full-scale.
  
  Drain Current
  The standard drain current constant is 1.21V/Amp. The standard calibration constant is 2,019. The output of the A-D convertor
  for a drain current of 1A is: 4096 * 1.21 / 5.0 = 991. The current calculation is therefore: 991 * 2,019 / 2,000,000 = 1.00A
  The over-current trip limit is the equivalent A-D convertor value scaled and corrected by the calibration constant. Therefore,
  in order to represent the drain current as a fraction of the current trip limit the bar graph meter reading is calculated:
	48 * Raw ADC / Overcurrent Trip. Turning things around, the id_const = cal.calval.overcurrent_trip / 48.

  Battery Voltage
  The display is effectively that of a suppressed-zero voltmeter. The minimum reading is the undervoltage trip limit,
  and the maximum is the over-voltage trip limit. The scale length is therefore: over-voltage - under-voltage
  The fractional length to display is thus: (actual voltage - undervoltage) / scale length
  and the number of bars is: 48 * fractional length
  The battery voltage is measured from a potential divider consisting of a 33K resistor in series with a 10K resistor.
  The fractional factor is therefore: 10 / (33 + 10) = 0.232558
  If we assume an undervoltage trip of 11V and an overvoltage trip of 15V, this gives a scale length of 4V, and as there
  are 8 divisions, this resolves to 0.5V/division. Using the default battery calibration multiplier of 5249, the
  undervoltage trip A-D value is therefore 2095, and the overvoltage A-D value is 2858.
  The scale length is therefore 2858 - 2095 = 763
  If the input voltage for example is 13V, this is an A-D value of 2477, batt_raw would therefore be 2477 * 5250 = 13004250
  Thus the first value is: v = (batt_raw - cal.calval.undervoltage_trip) * cal.calval.batt_mult / BATT_MULT
  The variable value would then be 48 * 2447 / 763 = 117456 / 763 = 153
  Taking account of the scale length, v = v /  (cal.calval.overvoltage_trip - cal.calval.undervoltage_trip)
  Since there are 6 divisions to each scale bar, this gives a fraction display of 153 / 6 = 25, or half scale.
  Note that there is a constant loop-invariant expression here: 48 * cal.calval.batt_mult / BATT_MULT,
  as well as: 48 / (cal.calval.overvoltage_trip - cal.calval.undervoltage_trip)
  It is pointless re-calculating this constant each time through the loop, so it is moved to the initialisation
  section where: batt_scale_factor = cal.calval.overvoltage_trip - cal.calval.undervoltage_trip
*/
void display_pwr_meter(void)
	{
	pwr_meter = (int)((double)out_pwr * power_bar_graph_const);		// Set global variable pwr_meter for TRX-2 protocol (serial_trx2.c)
	draw_s_meter(pwr_meter);
	}

// Display bar graph meter, S-meter when RX, PWR when TX, or alternative displays.
void disp_bar_meter(void)
	{
	if(PTT_IN)							// Receive Mode
		display_s_meter(eeprom.defval.S_Meter_Type);	// Display S-Meter
	else								// Transmit Mode
		{
		switch(sub_page0)
			{
			case 0: case 1: default:
				display_pwr_meter();	// Display Power-Meter
			break;

			case 2:						// Display SWR. SWR is 100 * SWR value, eg, 3:1 is a value of 300.
				if(swr < 100)			// This avoids garbage character in graphic display at
					draw_s_meter(0);	// first position when there is no reasonable output power.
				else
					draw_s_meter((int)int_round((long)(48 * (swr - 99)), (long)(cal.calval.swr_trip - 99)));
			break;

			case 3:						// Display Battery Voltage
				draw_s_meter((int)int_round((long)48L * (long)(batt_raw - cal.calval.undervoltage_trip), (long)batt_scale_factor));
			break;

			case 4:						// Display Drain Current
				draw_s_meter((int)int_round((long)id_raw_value, (long)id_const));
			break;
			}
		}
	}

// Measure & display Watt meter
void disp_fwd_pwr(void)
	{
	static long peak_pwr;

	if(PTT_IN == TRANSMIT)						// TX
		{
		if(out_pwr > peak_pwr)					// Fast attack
			{
			peak_pwr = out_pwr;					// Fast attack
			decay_counter = PEAK_SHOW_TIME;		// Reload decay timer (1,000mS)
			}
		else									// Slow release
			{									// No
			if(decay_counter == 0)
				{
				decay_counter = DECAY_TIME;		// Reload decay timer (20mS)
				peak_pwr -= (peak_pwr >> 1);
				}
			}

		disp_meter(((double)peak_pwr) * cal.calval.fwd_pwr_mult, WATTS);	// Scaling, P = ((ADC * ADC) * Calibration Factor / 1000000000
		}
	else										// RX
		{
		lcd_putst(Display_Watts);
		peak_pwr = 0;
		}
	}

// Measure & display Drain current meter
void disp_id(void)
	{
	static double drain_current;				// Initialised to zero by default.
	static int hold_time;
	double id;

	if(PTT_IN)									// TX ON ?
		{
		lcd_putst(Display_Amps);				// No, so display blank drain current
		drain_current = 0.0;
		id_raw_value = 0;
		}
	else										// Yes, so display drain current
		{
		id = (double)(id_raw_value) * (double)(cal.calval.id_mult);

		if(id > drain_current)
			{
			drain_current = id;
			hold_time = 125;
			}
		else
			{
			if(hold_time--); else drain_current = 0.0;
			}

		disp_meter((svc_flag) ? id : drain_current, AMPS);
		}
	}

/*
  Build main board & RF filter board (Only all band version) 74HC595 I/O latch data & send it
  D14 - D15 RF Attenuator selection
  Attenuation	D15		D14
		0		0		0
		1		0		1
		2		1		0
		3		1		1
  D8, D9, D10 RF filter select bits
  D7 - 3.5MHz/7.0MHz Filter Select bit (2-Band PA)
  D6 - Mic/Line I/P Select
  D5 - Noise Blanker On/Off (Not Used, always 0)
  D4 - Speech Processor On/Off
  D3 - AGC Speed, Fast/Slow
  D2 - Narrow SSB/CW Filter Select
  D1 - USB/LSB Sideband Select
  D0 - SSB/CW Mode Select
*/
void send_packet_to_74hc595(void)
	{
	unsigned int i;

	for(i = 0; i < RF_FILT_TABLE_LEN; i++)	// Scan filter & attenuator switch point table
		{
		if(freq_in_use < rf_filt_attn[i])
			break;		
		}

	rf_filt_idx = i;						// Set resolved index, 0 - 11

	if(tune_on && !adj_attn)				// Check for TUNE mode
		i = eeprom.defval.Tune_Attenuation;	// and get TUNE attenuation
	else									// Get preset band attenuation setting
		i = (unsigned int)cal.calval.rf_atn_bits[rf_filt_idx];

	i <<= 14;								// Shift attenuation setting up to bits 14 and 15,
	hc595 = rf_filter_bits[rf_filt_idx] | i;// and set filter select and attenuator bits.

	if(freq_in_use > M7_FILT)
		hc595 |= MHZ7;						// D7 - 3.5 / 7 MHZ filter select(for dual-band PA only)

	if(eeprom.defval.af_input == MIC)
		hc595 |= AF_INPUT;					// D6 - Microphone/Line input selection 0 = Line, 1 = Mic

//	if(eeprom.defval.nb != 0)
//			hc595 |= NB;					// D5 - Noise Blanker - No longer used - 5B4AIY

	if(eeprom.defval.proc_on_off == ON && eeprom.defval.af_input == MIC)
		hc595 |= PROC_ON;					// D4 - Speech processor On = 1, Off = 0

	if(eeprom.defval.agc_speed == FAST)
		hc595 |= FAST_AGC;					// D3 - AGC speed, Fast = 1, Slow = 0

	if(eeprom.defval.filter == NARROW)
		hc595 |= NAR;						// D2 - Narrow SSB/CW active band pass filter select
/*
  Mode: 0 = LSB, 1 = USB, 2 = CW (Normal), 3 = CWR (Reverse), 4 = Tune
  Truth Table for bits 0 and 1.
  Bit 0 = SSB/!CW Select, 1 = SSB, 0 = CW/Tune
  Bit 1 = USB/LSB Select 1 = LSB, 0 = USB

  Mode   SB    SSB/!CW
 ======================
       Bit 1    Bit 0
  LSB    1        1
  USB    0        1
  CW     0        0
  CWR    1        0
  Tune   1        0
 ======================
*/
	if(eeprom.defval.mode[vfo_idx] != USB && eeprom.defval.mode[vfo_idx] != CW)
		hc595 |= SB_SELECT;					// D1 - Sideband Select LSB = 1, USB = 0

	if(eeprom.defval.mode[vfo_idx] < 2)
		hc595 |= SSB_CW;					// D0 - SSB / !CW Select, USB/LSB = 1, CW/CWR/Tune = 0

	exchg_data_spi1(hc595);					// Start transmission to main board latch
	}

void get_pwr(void)
	{
	fwd_adval = convert_adc12(FWD_PWR);
	rev_adval = convert_adc12(REV_PWR);
	fwd_pwr = (long)fwd_adval * (long)fwd_adval;	// Calculate relative power levels (V^2)
	rev_pwr = (long)rev_adval * (long)rev_adval;
	out_pwr = fwd_pwr - rev_pwr;					// Output power = forward - reverse.
	}

void warning_beep(void)
	{
	if(decay_counter == 0)
		{
		beep(HZ1000, 25);
		decay_counter = 500;
		}
	}

void check_alarm(void)									// Executed each time through the main loop
	{
	static unsigned long avg_current;					// Initialised to zero by default.
	static int avg_ctr;									// Initialised to zero by default.

	if(cal.calval.alarm_flags & 0x01)					// If the over-current alarm is enabled...
		{
		if(avg_ctr)										// Get the next sample to average
			{
			id_raw_value = convert_adc12(ID_CUR);		// Save in global for use by other functions.
			avg_current += id_raw_value;
			avg_ctr--;									// Decrement the counter
			}
		else											// We have all the samples
			{
			avg_current /= OVERCURRENT_AVERAGE;			// Calculate the average current (64 samples)

			if((avg_current > cal.calval.overcurrent_trip) && !(alarm_flag & 0x01))
				{	// The current is over the limit, and the alarm flag has not been set
				alarm_flag |= 0x01;						// Set the alarm flag
				}

			avg_current = 0;							// Reset average current
			avg_ctr = OVERCURRENT_AVERAGE;				// Reset average counter
			}
		}

	batt_raw = convert_adc12(BATT_CH);					// Get battery voltage, and save in global for other functions.

	if(cal.calval.alarm_flags & 0x04)					// If the over-voltage alarm is enabled...
		{
		if((batt_raw > cal.calval.overvoltage_trip) && !(alarm_flag & 0x04))
			{	// The voltage is over the limit, and the alarm flag has not been set
			alarm_flag |= 0x04;							// Set the alarm flag
			}
		}
	
	if((batt_raw < low_voltage_warning) && !(alarm_flag & 0x10))	// If the low-voltage warning alarm is enabled...
		{		// Voltage is below the low voltage warning, and the alarm flag has not been set
		alarm_flag |= 0x10;								// Set the alarm flag and display message flag
		}

	if(cal.calval.alarm_flags & 0x08)					// If the under-voltage alarm is enabled...
		{
		if((batt_raw < cal.calval.undervoltage_trip) && !(alarm_flag & 0x08))
			{	// The voltage is below the cut-off limit, and the alarm flag has not been set
			alarm_flag |= 0x08;							// Set the alarm flag and display message flag
			}
		}

	if(alarm_flag)
		{
		lcd_mode = ALARM_DISPLAY;	// Force alarms to be displayed
		mute = AUDIO_ON;			// Turn audio on
		}
	}

// Function for Multiband PA input RF attenuator adjust
void adjust_rf_attenuator(void)
	{
	int temp_sub_page, atten;

	adj_attn = TRUE;								// This forces the use of the true attenuator setting
	temp_sub_page = sub_page0;						// Save the current display page,
	sub_page0 = 1;									// and switch to the RF output power display.

	do	{
		check_alarm();
		get_pwr();
		lcd_cmd(LINE1);
		disp_bar_meter();							// Display PWR meter bar graph.
		disp_meter((double)out_pwr * (double)cal.calval.fwd_pwr_mult, WATTS);	// Scaling: P = ((ADC * ADC) * Calibration Factor / 1000000000
		lcd_cmd(LINE2);
		sprintf(lcdpbuff, Display_Attenuator, -cal.calval.rf_atn_bits[rf_filt_idx]);
		lcd_putst(lcdpbuff);
// If you attempt to use: &cal.calval.rf_atn_bits[rf_filt_idx] you will provoke a compiler 'Incompatible pointer type'
// error message because cal.calval.rf_atn_bits[] is an array of characters, not integers. Thus, to avoid this error
// the value is equated to an integer, the range is checked, and the result returned to the array. A.Ryan - 30/JAN/2014
		atten = cal.calval.rf_atn_bits[rf_filt_idx] - encoder_get_coarse(ENCODER_SLOW);
		max_min(&atten, MAX_ATTEN, MIN_ATTEN);
		cal.calval.rf_atn_bits[rf_filt_idx] = atten;
		send_packet_to_74hc595();
		warning_beep();
		} while(!PTT_IN_IO && !alarm_flag);			// Stay here while PTT switch is pressed, and there are no alarms.

// Prepare for exit
	if(alarm_flag) clear_lcd();
	else save_calval(1);							// Save calibration values on exit, if no alarms.

	reset_encoder();								// clear the encoder,
	sub_page0 = temp_sub_page;						// and restore prevous display page.
	adj_attn = FALSE;								// This forces the use of the preset attenuator setting
	}

// LCD back light functions
void bl_on(void)
	{
	set_pwm3_dac(eeprom.defval.back_light);			// Turn backlight ON
	bl_timer = eeprom.defval.back_light_timer * 5;	// Load timer
	}

void set_bit(int index)
	{
	unsigned long mask = 1;

	eeprom.defval.VFO_Lock |= (mask << index);
	}

void clr_bit(int index)
	{
	unsigned long mask = 1;

	eeprom.defval.VFO_Lock &= ~(mask << index);
	}

// Handle PWR button. VFO lock, Shut down the system or RF attenuator adjust if TX && TUNE, improved sequence 03,07.2008
void power_off(void)
	{
	int print_flag = FALSE;

	if(alarm_flag)
		{
		long_push = MEDIUM_PUSH;					// Prepare for emergency shutdown
		beep(HZ466_85, cal.calval.beep_len);		// Short beep for push noticed
		loop_ctr = 0;								// Reset the alarm blink counter

		if(alarm_flag & 0x01)						// Is it Over-current?
			{
			alarm_flag &= 0x0E;						// Try to clear over-current alarm
			goto Continue;
			}

		if(alarm_flag & 0x02)						// Is it High SWR?
			{
			alarm_flag &= 0x0D;
			goto Continue;
			}

		if(alarm_flag & 0x04)						// Is it Over-Voltage?
			{
			alarm_flag &= 0x0B;
			goto Continue;
			}

		if(alarm_flag & 0x08)						// Is it Under-Voltage?
			{
			alarm_flag &= 0x07;
			goto Continue;
			}

		if(alarm_flag & 0x10)						// Is it Under-Voltage warning?
			{
			alarm_flag &= 0x0F;
			low_voltage_warning = 0;				// Reset trigger - one-time alarm!
			goto Continue;
			}

	Continue:										// I am not dogmatic about the non-use of GOTO

		while(PWR_SW)
			{
			if(long_push == 0)
				{
				display_screen(Emergency, Shutdown);
				stop_scaf_clock();

				while(PWR_SW);						// Wait for PWR button to be released

				PWR_ON = OFF;
				while(TRUE);
				}
			}

		ms_delay(BUTTON_DEBOUNCE);
		return;
		}

	if(tune_on)										// PWR switch in Tune mode takes us to
		adjust_rf_attenuator();						// PA RF attenuator (Power level) adjustment
	else
		{
		vfo_lock ^= 1;								// VFO tune lock flip-flop

		if(vfo_lock)
			{
			beep(HZ466_85, cal.calval.beep_len);	// Short beep for push noticed, Locked
			set_bit(vfo_idx);
			}
		else
			{
			beep(HZ587_31, cal.calval.beep_len);	// Short beep for push noticed, Unlocked
			clr_bit(vfo_idx);
			}

		ms_delay(250);
		long_push = POWER_OFF;

		while(PWR_SW)
			{
			if(long_push == 0)								// Commence power down sequence
				{
				beep(HZ392_01, cal.calval.beep_len * 2);	// Play low sound when ready for power down
				ms_delay(cal.calval.beep_len * 2);			// Let the tone finish
				vfo_lock ^= 1;								// This reverses the spurious VFO Lock/Unlock
															// that occured when the PWR button was pressed
				if(vfo_lock)
					set_bit(vfo_idx);
				else
					clr_bit(vfo_idx);

				if(power_down_mode == SAVE_STATE)			// Save current User Configuration
					{
					save_defaults();
					display_screen(Configuration_Msg, saved_restored[1]);
					print_flag = TRUE;
					ms_delay(850);
					}

				if(changed_flag) save_defaults();			// Save any unsaved Rapid Bandswitch User or VFO memory changes.

				if(power_down_mode == PROMPTED)				// Prompted Power OFF
					print_flag = save_settings(0, 2);		// Prompt to save current configuration settings

				if((eeprom.defval.serial_mode == SERIAL_TEST) && print_flag)
					{
					printf(Settings_Saved_EEPROM);
					ms_delay(5);							// Let the printout finish
					}

				stop_scaf_clock();							// Block audio path, reduce power off noise
				clear_lcd();
				bl_off();

				while(PWR_SW);								// Wait while power switch is still pressed

				PWR_ON = OFF;								// Release the power latch
				while(TRUE);								// Final wait
				}
			}
		}
	}

// RS-232 loop back test - Press and hold DISPLAY and then power on until prompt is displayed
void rs232_test(void)
	{
	unsigned char c;

	display_screen(Loop_Back_Test, Awaiting_Data);
	printf(RS232_Test_Exit);

	do	{
		if(kbhit())
			{
			c = getch();
			beep(HZ698_45, SHORT_BEEP);		// Give beep tone
			putch(c);						// Echo character to terminal
			sprintf(lcdpbuff, Display_RX_Data, c, c);
			lcd_cmd(LINE2);			
			lcd_putst(lcdpbuff);
			}

		if(PWR_SW)							// Exit test
			{
			printf(Test_Terminated);
			break;
			}
		} while(TRUE);
	}

/*
 Display User Configuration Settings. Note: This only displays the current settings,
 the adjustment is performed in the User Configuration section.
 Many of the top-line prompts were the same, so it made sense to try and re-use these common prompts.
 Consequently, the original version used a long SWITCH-CASE using the page variable to set an index into
 the user_cfg_prompts[] array. This compiler does not implement the SWITCH-CASE very efficiently, and
 this version now adds an array of indices, prompt_index[], eliminating the SWITCH-CASE, and this array
 contains the appropriate index into the user_cfg_prompts[] array. Whilst this added 28 bytes to the data
 area, it saved 21 bytes of code space and was more elegant. A.Ryan - 2/FEB/2014
*/
int display_settings(int page, int adjust)
	{
	int i;

	i = page;

	if(adjust)				// Change page if in fast page select mode
		{
		page += encoder_get_coarse(ENCODER_MEDIUM);
		max_min(&page, MAX_SUB_PAGE1, 0);
		}

	if((eeprom.defval.keyer < STRAIGHT_KEY) && (page == 5))			// If keyer is in an Iambic or Vibroplex mode,
		{
		if(i <= page)		// then skip to next page,
			page++;
		else
			page--;			// or skip to previous page.
		}

	sprintf(lcdpbuff, _16s, user_cfg_prompts[prompt_index[page]]);	// Select appropriate prompt message,
	lcd_cmd(LINE1);			// position the cursor to the start of line 1,
	lcd_putst(lcdpbuff);	// and display the prompt.
	display_pages[page]();	// Select appropriate adjustment page,
	lcd_cmd(LINE2);			// position the cursor to the start of line 2,
	lcd_putst(lcdpbuff);	// and display the return message.
	return page;
	}

/*
 The design of the alarm blink feature is such as to always start with a synchronised blink. Although the
 mSec counter could be used, you could not guarantee that the message would always be displayed for the
 full time as soon as an alarm was detected. There would also be unsynchronised beeps. In this implementation,
 when an alarm is detected the cursor is set to line 1, and the loop_ctr is checked to see if it zero.
 The normal state is zero, but when an alarm is detected, then it get set to 900, a beep is given, and then
 the alarm variable checked to see which alarm flag is set, in order of priority. The appropriate message
 is displayed, and the function returns.

 On the next iteration of the main loop, although the alarm flag is still set, the loop_ctr is decrementing, and
 so the function returns without disturbing the display. Eventually the loop_ctr decrements to a value of 450, and
 then the display blank flag is set. On the next iteration of the main loop this is detected, and the top line of the
 display is over-written with blanks, and a beep sounds. When the function exits the display blanks flag is cleared.
 Once again, although the alarm flag is set, nothing happens to the display. Eventually the loop_ctr decrements to zero
 and the process repeats. If the PWR button is pressed, the alarm flag may be cleared, if so, on the next iteraion of
 the main loop this is detected, the lcd_mode is returned to normal, and the display returns to its normal mode.
 If the alarm flag cannot be cleared, then by holding the PWR button down an immediate shutdown is forced.
*/
void display_alarm(void)
	{
	if(!alarm_flag)
		lcd_mode = NORMAL_DISPLAY;							// If the alarms are clear, restore normal display
	else
		{
		bl_on();											// Force backlight on
		lcd_cmd(LINE1);

		if(loop_ctr == 0)	// The loop counter eventually gets decremented to zero in the 1mS timer interrupt
			{
			loop_ctr = LOOP_COUNT;							// Reset the blink counter
			beep(4700, SHORT_BEEP);

			if(alarm_flag & 0x01)
				{
				lcd_putst(Overcurrent_Alm);
				return;
				}

			if(alarm_flag & 0x02)
				{
				lcd_putst(High_SWR_Alm);
				return;
				}
				
			if(alarm_flag & 0x04)
				{
				lcd_putst(Over_Voltage_Alm);
				return;
				}

			if(alarm_flag & 0x08)
				{
				lcd_putst(Under_Voltage_Alm);
				return;
				}

			if(alarm_flag & 0x10)
				{
				lcd_putst(Low_Battery_Alm);
				return;
				}
			}

		if(alarm_flag & 0x20)							// Display blank flag set
			{
			lcd_spc(16);								// Display blanks
			alarm_flag &= 0x1F;							// Clear Display Blank flag
			beep(4700, SHORT_BEEP);
			}
		}
	}

long get_average_pwr(void)
	{
	long favg_pwr = 0, ravg_pwr = 0;
	int i;

	for(i = 0; i < 10; i++)
		{
		ms_delay(75);
		get_pwr();
		favg_pwr += fwd_adval;
		ravg_pwr += rev_adval;
		lcd_cmd(LINE1);
		display_pwr_meter();		// Display S-meter or PWR meter bar graph
		disp_meter((double)fwd_pwr * (double)cal.calval.fwd_pwr_mult, WATTS);	// Scaling, P = ((ADC * ADC) * Calibration Factor / 1000000000
		check_alarm();

		if(calc_swr(fwd_adval, rev_adval) > cal.calval.swr_trip) alarm_flag |= 2;
		}

	favg_pwr /= 10;
	ravg_pwr /= 10;
	favg_pwr *= favg_pwr;
	ravg_pwr *= ravg_pwr;
	favg_pwr -= ravg_pwr;
	return favg_pwr;
	}

int get_atten(double ref, double measured)
	{
/*
  The actual attenuation values taking into account nominal resistor values are:
  -1dB attenuator: -0.936dB R1/R3 = 820R, R2 = 4R7 
  -2dB attenuator: -1.960dB R1/R3 = 470R, R2 = 12R
  -3dB attenuator: -2.982dB R1/R3 = 270R, R2 = 16R
*/
	double ratio;

	ratio = measured / ref;
	ratio = 10 * log10(ratio);
	printf(Relative_Power, ratio);

	if(ratio >= 1.8) return 3;

	if(ratio >= 1.3) return 2;

	if(ratio >= 0.8) return 1;

	return 0;
	}

void auto_equalise(void)
	{
	int save_atten[9], new_atten[9], save_tune_atten, save_mode, i, complete = FALSE;
	long measured_power[9], reference_pwr;

	save_tune_atten = eeprom.defval.Tune_Attenuation;	// Save current Tune Attenuation setting
	eeprom.defval.Tune_Attenuation = 1;					// Set attenuator to -1dB
	save_mode = eeprom.defval.mode[vfo_idx];			// Save current VFO's mode
	eeprom.defval.mode[vfo_idx] = TUNE;					// Force TUNE mode

	for(i = 0; i < 9; i++)								// Save current amplifier equalisation settings
		{
		save_atten[i] = cal.calval.rf_atn_bits[i + 3];
		cal.calval.rf_atn_bits[i + 3] = 0;
		}

	printf(Press_Hold);

	while(PTT_IN_IO);			// Wait for PTT switch closure

	printf(Equalising);

	while(!PTT_IN_IO)
		{
		KEY = ON;
		tune_on = ON;

		for(i = 9; i > 0; i--)
			{
			freq_in_use = amateur_bands[i - 1];				// Select next frequency
			lcd_cmd(LINE2);
			disp_freq(freq_in_use);							// Display current frequency
			dds_load(freq_in_use);							// Load the synthesiser with the new frequency
			send_packet_to_74hc595();						// and select the new RF filter
			measured_power[i - 1] = get_average_pwr();		// save the measured power
			printf(Averaged_Pwr, band[i], ((double)measured_power[i - 1] * (double)cal.calval.fwd_pwr_mult) / 1000000000.0);

			if(PTT_IN_IO || alarm_flag) goto Abort;			// If PTT switch is released, or there is an alarm, abort
			}

		printf(New_Line);
// Now get the minimum power to act as the reference
		
		reference_pwr = 20000000L;	// Set to some impossibly high figure

		for(i = 0; i < 9; i++)
			{
			if(measured_power[i] < reference_pwr)
				reference_pwr = measured_power[i];
			}

// Now calculate the attenuator settings
		for(i = 0; i < 9; i++)
			{
			new_atten[i] = get_atten((double)reference_pwr, (double)measured_power[i]);
			printf(Band_Atten, band[i + 1], -new_atten[i]);
			}

		complete = TRUE;
		printf(Release_PTT);

		while(!PTT_IN_IO);		// Wait for PTT switch to be released

		ms_delay(500);			// Take care of contact bounce, otherwise the whole thing repeats
		}

	Abort:						// I'm not dogmatic about not using GOTO if it's sensible, as here.

		printf(New_Line);

		for(i = 0; i < 9; i++)
			{
			if(complete)
				cal.calval.rf_atn_bits[i + 3] = new_atten[i];
			else
				cal.calval.rf_atn_bits[i + 3] = save_atten[i];
			}

		if(complete)
			{
			save_calval(0);
			printf(Equalisation_Complete);
			}
		else
			printf(Equalisation_Aborted);

// Restore the TUNE mode attenuator setting, and original VFO mode
	eeprom.defval.Tune_Attenuation = save_tune_atten;		// Restore original Tune Attenuation setting
	eeprom.defval.mode[vfo_idx] = save_mode;				// Restore original VFO's mode
	KEY = OFF;
	tune_on = OFF;
	}

// Adjust VFO frequency to the nearest decade for the frequency step in use
void adjust_vfo(void)
	{
	long f, r, step;

	if(!eeprom.defval.VFO_Adjust || !eeprom.defval.step_idx)	// If turned off, or step is 1Hz, then return.
		return;

	if(eeprom.defval.VFO_Adjust >= eeprom.defval.step_idx)		// If VFO Adjust >= Current Adjustment Setting
		{
		step = f_steps[eeprom.defval.step_idx];
		f = eeprom.defval.vfo[vfo_idx];
		r = f % step;
		f -= r;								// First round down.

		if(r >= (step >> 1))				// If the remainder is greater than or equal to half the step value,
			f += step;						// round up one step.

		eeprom.defval.vfo[vfo_idx] = f;		// Now set the VFO to nearest step decade.
		}
	}

int check_bit(int index)
	{
	unsigned long mask = 1;

	mask <<= index;
	return (eeprom.defval.VFO_Lock & mask) ? TRUE : FALSE;
	}

long get_rit(void)
	{
	int i;

	i = convert_adc12(RIT_POT);			// Get RIT pot value

	if(abs(old_ADC - i) > 1)
		old_ADC = i;
	else
		i = old_ADC;					// This eliminates the +/-1 Hz jitter.

	if(i >= (2048 + RIT_DEAD_ZONE))		// Blank dead zone (RIT_DEAD_ZONE = 47, defined in juma-trx2.h)
		i -=(2048 + RIT_DEAD_ZONE);		// Positive side
	else
		{
		if(i <= (2048 - RIT_DEAD_ZONE))	// Negative side
			i -=(2048 - RIT_DEAD_ZONE);
		else
			i = 0;						// Dead zone
		}

	return (i / 2);						// In decimal frequency system, we divide the A-D value by 2
	}

void display_hdr(void)
	{
	printf(firmware, Display_JUMA_TRX2, ver, date);
	printf(copyright);
	printf(added_features);
	}

void get_one_zero(int *value)
	{
	int x;

	x = encoder_get_coarse(ENCODER_SLOW);
//	*value = (x > 0) ? 1 : (x < 0) ? 0 : *value;	// This uses 16 bytes more code! A.Ryan - 5B4AIY
	if(x > 0) *value = 1;
	if(x < 0) *value = 0;
	}

void get_value(int *value, int speed, int min, int max)
	{
	if(svc_flag && latch) return;

	*value += encoder_get_coarse(speed);
	max_min(value, max, min);
	}

void set_decade_value(int *value, int min, int max)
	{
	if(svc_flag && latch) return;

	*value += encoder_get_coarse(ENCODER_MEDIUM) * 10;
	max_min(value, max, min);
	}

void rs232_2(void)
	{
	serial_trx2(!PTT_OUT & PTT_IN);		// Flag indicates the freqency has changed. PTT_OUT 1 = TX, 0 = RX, PTT_IN 0 = TX, 1 = RX
	}

void (*port_mode[])() = {				// Indexed by: eeprom.defval.serial_mode
						serial_yaesu,	// Yaesu CAT
						serial_kenwood,	// Kenwood CAT
						rs232_2,		// TRX-2 protocol
						serial_test,	// Test mode
						serial_ext_kb	// Voice Memory
						};
/*
  One of the features of elegant well-structured code is the concept of burying the complexity or details.
  The original code used several very long and complex switch/case constructs to deal with setting the user
  configuration and calibration setup. Whilst there is nothing inherently 'wrong' with this approach, it did
  lead to a very complex source code layout where one could easily get lost amongst the multitude of nested
  loops and braces. Changes were difficult to accomplish, and it was very easy to place a containing brace in
  the wrong position, so that although the code would compile and link without any apparent error, there could
  be really serious or subtle bugs in the executable code. To avoid this and to simplify the high-level source
  code I decided to restructure these long switch/case constructs into a simpler form. The following sections
  of code accomplish this by placing the various display pages etc, into small functions, and then collecting the
  functions into an array and finally executing the desired function using a simple index that was the original
  code's switch variable. Although this has increased the size of the executable code module the overall
  simplification was well worth the effort. The increase in code size is largely as a result of this processor's
  lack of a complex indexed call instruction which is common amongst CISC processors. As a consequence it takes
  a few more machine cycles to actually calculate the desired execution address. - Adrian Ryan - 5B4AIY - 22/MAY/2012
*/
// AGC speed - Fast/Slow
void sp_0(void)
	{
	get_one_zero(&eeprom.defval.agc_speed);
	}

// Adjust selected filter
void sp_1(void)
	{
	eeprom.defval.scaf_filters[eeprom.defval.filter] -= encoder_get_coarse(ENCODER_MEDIUM);		// Add changes
	max_min(&eeprom.defval.scaf_filters[eeprom.defval.filter], scaf_max[eeprom.defval.filter], scaf_min[eeprom.defval.filter]);
	}

// PROC On / Off
void sp_2(void)
	{
	get_one_zero(&eeprom.defval.proc_on_off);
	}

// AF Input Source - Mic/Line
void sp_3(void)
	{
	get_one_zero(&eeprom.defval.af_input);
	}

// Keyer operating mode - Dot Priority, Iambic A, Iambic B, Vibroplex, or Straight
void sp_4(void)
	{
	get_value(&eeprom.defval.keyer, ENCODER_SLOW, DOT_PRIORITY, STRAIGHT_KEY);
	}

// Straight Key CW Release Time Minimum = 200mS, Maximum = 1.5 Seconds
void sp_5(void)
	{
	set_decade_value(&eeprom.defval.CW_Release, MIN_CW_RELEASE, MAX_CW_RELEASE);
	}

// Set CW Shift & Side Tone Frequency, Minimum = 300 Hz, Maximum = 1 kHz
void sp_6(void)
	{
	get_value(&eeprom.defval.cw_shift, ENCODER_SLOW, MIN_CW_SHIFT, MAX_CW_SHIFT);
	eeprom.defval.cw_sidetone = FCY / (2 * eeprom.defval.cw_shift);
	}

// CW speed pot = CW speed / Squelch-A / Squelch-B
void sp_7(void)
	{
	get_value(&eeprom.defval.sq, ENCODER_SLOW, 0, 2);
	}

// CW Keyer Speed Preset
void sp_8(void)
	{
	int x;

	x = encoder_get_coarse(ENCODER_MEDIUM);

	if(x > 0) eeprom.defval.CW_Speed |= LOCKED;		// Preset speed locked

	if(x < 0) eeprom.defval.CW_Speed &= UNLOCKED;	// Preset speed unlocked
	}

// CW/CWR Swap Enable/Disable
void sp_9(void)
	{
	get_one_zero(&eeprom.defval.CW_CWR_Swap);
	}

// LCD back light intensity				
void sp_10(void)
	{
	set_decade_value(&eeprom.defval.back_light, MIN_BL, MAX_BL);
	set_pwm3_dac(eeprom.defval.back_light);
	}

// LCD Contrast
void sp_11(void)
	{
	set_decade_value(&eeprom.defval.contrast, MIN_CONTRAST, MAX_CONTRAST);
	set_pwm4_dac(eeprom.defval.contrast);
	}

// LCD back light timer
void sp_12(void)
	{
	get_value(&eeprom.defval.back_light_timer, ENCODER_SLOW, 0, MAX_BLT);
	}

// Serial I/O mode/protocol select.
// Voice Memory mode only selectable if the module is present. A.Ryan - 5B4AIY - 17/AUG/2013
void sp_13(void)
	{
	ClearUSART1queue();
	ClearUSART2queue();						// This will hopefully eliminate the spurious VM Erase bug. A.Ryan - 01/MAR/2014
	get_value(&eeprom.defval.serial_mode, ENCODER_SLOW, 0, 3 + isd_1700_installed);

	if(eeprom.defval.serial_mode < 2)		// If either CAT mode selected
		{
		eeprom.defval.max_vfo = SPLIT_MODE;	// Force VFO mode to A/B + Split
		eeprom.defval.vfo_sel = VFO_A;		// Force active VFO as VFO-A
		}

	RS_232_Mode = port_mode[eeprom.defval.serial_mode];
	}

// Select user UART1 Baud rate
void sp_14(void)
	{
	get_value(&eeprom.defval.br, ENCODER_SLOW, 0, 7);
	SetUSART1baud(eeprom.defval.br);		// Set new Baud rate
	}

// Select VFO mode / Number of memories in use
void sp_15(void)
	{
	if(eeprom.defval.serial_mode > 1)		// If neither CAT mode is selected
		{
		get_value(&eeprom.defval.max_vfo, ENCODER_SLOW, 2, 26);
		eeprom.defval.vfo_sel = 0;			// Reset to VFOA
		}
	}

// Select type of S-Meter display, graphic (original) or numeric
void sp_16(void)
	{
	get_one_zero(&eeprom.defval.S_Meter_Type);
	}

// Numeric S-Meter Peak Hold Time Minimum = 250mS, Maximum = 5 Secs.
void sp_17(void)
	{
	set_decade_value(&eeprom.defval.Hold_Time, MIN_S_METER_HOLD, MAX_S_METER_HOLD);
	}

// Enable / Disable out-of-band checks for transmitter inhibit
void sp_18(void)
	{
	get_one_zero(&eeprom.defval.OOB_Check);
	}

// Auto Sideband Select Enable/Disable
void sp_19(void)
	{
	get_one_zero(&eeprom.defval.Auto_SB);
	}

// Frequency Display Old/New select
void sp_20(void)
	{
	get_value(&eeprom.defval.Freq_Display, ENCODER_SLOW, 0, 3);
	}

// Band Switch Mode Select 0 = Default, 1 = User Frequencies, 2 = User Frequencies (Locked)
void sp_21(void)
	{
	get_value(&eeprom.defval.Band_SW, ENCODER_SLOW, 0, 2);
	}

// Choose whether to include the RIT frequency with the displayed frequency
void sp_22(void)
	{
	get_one_zero(&eeprom.defval.RIT_Include);
	}

// Select attenuation in TUNE mode
void sp_23(void)
	{
	get_value(&eeprom.defval.Tune_Attenuation, ENCODER_SLOW, 0, 3);
	}

// Select VFO decade adjustment
void sp_24(void)
	{
	get_value(&eeprom.defval.VFO_Adjust, ENCODER_SLOW, 0, 3);
	}

// Graphic Scale Display select, Original, Large Markers, Small Markers
void sp_25(void)
	{
	get_value(&eeprom.defval.Display_Type, ENCODER_SLOW, ORIGINAL_MARKERS, SMALL_MARKERS);
	}

// Noise blanker datapoint now used for Fast Power OFF flag
void sp_26(void)
	{
	get_value(&eeprom.defval.nb, ENCODER_SLOW, IMMEDIATE, SAVE_STATE);

	if(power_down_mode == IMMEDIATE && display_page == LAST_PAGE) display_page = MODE_BW;
	}

// Start-Up Page Selection
void sp_27(void)
	{
	get_value(&display_page, ENCODER_SLOW, MODE_BW, (power_down_mode == IMMEDIATE) ? PA_CURRENT : LAST_PAGE);
	}

void (*set_value[])(void) = {
							sp_0,
							sp_1,
							sp_2,
							sp_3,
							sp_4,
							sp_5,
							sp_6,
							sp_7,
							sp_8,
							sp_9,
							sp_10,
							sp_11,
							sp_12,
							sp_13,
							sp_14,
							sp_15,
							sp_16,
							sp_17,
							sp_18,
							sp_19,
							sp_20,
							sp_21,
							sp_22,
							sp_23,
							sp_24,
							sp_25,
							sp_26,
							sp_27
							};

void dm_0(void)
	{
// Now display the current mode or blank
	if(dsp2_ctr == 2 || dsp2_ctr == 0)		// If counter is non-blank, display the mode
		lcd_putst(mode_msg[eeprom.defval.mode[vfo_idx]]);	// Display mode (LSB/USB/CW/CWR/Tune)
	else									// Else display a blank, 4 spaces for the Tune mode, or 3 for the others
		lcd_spc(4);
// Display filter in use
	if(PTT_IN)								// TX on ?
		{									// No, so display filter widths or blanks
		if(dsp1_ctr == 2 || dsp1_ctr == 0)
			lcd_putst(filter_width[eeprom.defval.filter]);
		else
			lcd_spc(3);
		}
	else
		lcd_putst(Display_TX);				// Yes, so display TX message instead.
	}

void dm_2(void)
	{
	(swr == 0) ? lcd_putst(Display_SWR) : disp_meter((double)swr, SWR);
	}

void dm_3(void)
	{
	double v;

	v = (double)batt_raw * (double)(cal.calval.batt_mult);
	batt_mean_val = (batt_mean_val * 0.9) + (v * 0.1);		// 10-sample running average.
	disp_meter(batt_mean_val, VOLTS);						// New version - 5B4AIY
	}

void (*dsp_mode[])(void) = {			// Indexed by sub_page0
						dm_0,			// 0 - Normal Display S-Meter(Rx)/Analog Power Meter(Tx), Mode, Filter
						disp_fwd_pwr,	// 1 - S-Meter(Rx)/Analog Power Meter(Tx), Digital Wattmeter
						dm_2,			// 2 - S-Meter(Rx)/Analog SWR(Tx), Digital SWR (disp_meter())
						dm_3,			// 3 - S-Meter(Rx)/Analog Voltmeter(Tx), Digital Voltmeter (disp_meter())
						disp_id			// 4 - S-Meter(Rx)/Analog Ammeter(Tx), Digital Ammeter
						};

void lcd_0(void)								// Normal display mode, cursor is already at the start of Line 1
	{
	if(eeprom.defval.filter == TX_FILTER)		// If the last filter adjustment was the TX filter,
		eeprom.defval.filter = WIDE;			// restore normal RX filter display
		
	disp_bar_meter();							// Display S-meter or PWR meter bar graph
	dsp_mode[sub_page0]();						// Display sub page (Normal/Power/SWR/Voltage/Current)
	lcd_cmd(LINE2);								// Second display line

// Test, RIT display in place of VFO info
// Freq display visible always
	if(eeprom.defval.RIT_Include)
		disp_freq(freq_in_use);
	else
		disp_freq(eeprom.defval.vfo[vfo_idx]);

	lcdoutch(' ');

	if(eeprom.defval.rit_on == ON && PTT_IN == RECEIVE)	// RIT ON and RX
		{
		disp_rit(rit);								// Special RIT display format
		}
	else
		{
// Display VFO
		if(split_flag)
			lcd_putst(Display_Splt);
		else
			{
			lcd_putst(Display_VFO);
			lcdoutch(eeprom.defval.vfo_sel + 'A');	// Display VFO letter
			}

// Display tuning step
		lcdoutch(lock_symbol[vfo_lock]);
		lcdoutch(tuning_step[eeprom.defval.step_idx]);
		}
	}

void lcd_1(void)
	{
	bl_on();										// Set / Keep LCD back light on
	set_value[sub_page1]();							// Set parameter value
	sub_page1 = display_settings(sub_page1, FALSE);	// Display new settings
	}

void (*lcd_dsp[])(void) = {				// Indexed by lcd_mode
						lcd_0,			// 0 - Normal Display
						lcd_1,			// 1 - User Configuration Display
						display_alarm	// 2 - Alarm Display
					};

// main __________________________________________________________________________________

int main(void)
	{
	long offset;
	int src_vfo, dst_vfo, audio_src, rapid_band_switch_flag, old_cw_pot = 0;
	unsigned int y, w, loop_delay_time;
	char c = FALSE;					// Do not use for anything else! Reserved exclusively for MODE button as VFO copy flag and direction arrow character!

// Set general I/O
	TRISA = INIT_TRISA;				// SWITCHES, PTT
	PORTA = INIT_PORTA;

	TRISB = INIT_TRISB;				// ADC inputs, TXEN, RXEN, General I/O
	PORTB = INIT_PORTB;

	TRISC = INIT_TRISC;				// Sideband select & switches
	PORTC = INIT_PORTC;

	TRISD = INIT_TRISD;				// LEDs & LCD pins = output
	PORTD = INIT_PORTD;

	TRISF = INIT_TRISF;				// DDS controls
	PORTF = INIT_PORTF;

	TRISG = INIT_TRISG;				// DDS controls, EEPROM, DScard signals
	PORTG = INIT_PORTG;

// Power switch
	PWR_ON = ON;							// Secure power ON

// Start PWM system
	IPC0 = 0x5444;							// Set tone generator (TMR1) priority higher than others
	init_timers_pwm();						// Setup PWM & Tone Generators

// Start DDS
	dds_init();								// Init AD9850
	dds_load(0);							// Do first dummy load
	dds_load(DEFAULT_FREQ_A);				// Dial in something, start mixer, seek operating point

// CW
	cw_period = 50;							// Set something before IRQ system starts
	fcy = FCY / 1000;						// Clock frequency (kHz)
	glitch_time = 100;						// Set startup glitch removal timer value - 5B4AIY

// Initialise Main Board SPI traffic
	init_spi1();
	exchg_data_spi1(hc595);

// Start LCD
	initlcd();								// Init LCD

// Put hello messages to LCD
	clear_lcd();
	lcd_putst(Display_JUMA_TRX2);			// Say hello
	lcd_putst(ver);							// Display version
	lcd_cmd(LINE2);
	lcd_putst(Display_OH2NLT_OH7SV);
	ms_delay(1000);

// Start UART & print hello messages
	InitUSART1();							// Wake up serial interface
	InitUSART2();							// Wake up secondary serial interface

// Clear UART input buffers before start
	ClearUSART1queue();
	ClearUSART2queue();

// Start ADC
	init_adc12();							// Setup A/D converter, must be done before Keyer is started, RB I/O bits

// Read calibration & default values from EEPROM
	y = read_calval();
	w = read_defaults();

	if(y != 0 || w != 0)					// Compare checksums
		{
		display_screen(Display_ChkSumErr, Display_Load_Defaults);
		set_factory_defaults();				// EEPROM empty or corrupted, set factory defaults
		ms_delay(1000);
		}

// Init ISD1700 voice memory if present.
// If not checked here, then the audio channels are not set, and the Service Mode beep is inaudible.
	init_isd1700();

// Set LCD contrast & back light with EEPROM data
	set_pwm4_dac(eeprom.defval.contrast);
	set_pwm3_dac(eeprom.defval.back_light);

// LCD back light timer
	bl_on();								// Set LCD back light

// Start encoder
	encoder_init();							// Init encoder, enable INT4

// Open audio channel and give start up notification 
	init_scaf_clock();						// Open audio channel, start SCAF clocking
	ms_delay(50);
	beep(HZ698_45, SHORT_BEEP * 2);			// Give start up beep, always

// Do calibration functions if service mode start
// Check for Service Mode
	long_push = VERY_LONG_PUSH;				// Set long push timer

	while(PWR_SW)
		{
		if(long_push == 0)
			{
			init_scaf_clock();				// Open audio channel, start SCAF clocking
			beep(HZ698_45, SHORT_BEEP * 4);	// Give service beep
			lcd_cmd(LINE2);	
			lcd_putst(Calibration_Mode);

			while(PWR_SW);					// Show message as long as Power switch is pressed

			ms_delay(750);
			display_screen(Display_Next, Display_Save);
			ms_delay(2000);					// and wait.
			svc_flag = TRUE;
			service();
			}
		}

	svc_flag = FALSE;

// Check if factory defaults requested
	if(_MODE) save_settings(1, 4);				// If MODE button pressed, restore and save factory default calibration and configuration settings.

// Set last used frequency
	if(eeprom.defval.vfo_sel == SPLIT_MODE)
		vfo_idx = VFO_A;						// If in split mode, set to VFO-A
	else
		vfo_idx = eeprom.defval.vfo_sel;		// Copy VFO selection for processing

	freq_in_use = eeprom.defval.vfo[vfo_idx];	// Set last used frequency
	dds_load(freq_in_use);						// Load DDS

// Set user Baud rate
	SetUSART1baud(eeprom.defval.br);

	if(eeprom.defval.serial_mode == SERIAL_TEST)// Test Mode selected, so display boot messages
		{
		display_hdr();
		printf(baud_rate_msg, br_txt[eeprom.defval.br]);
		printf(Display_System_Clk, fcy);
		printf(VM_Option_Found, Not_Installed[isd_1700_installed]);
		printf(Display_EEPROM_Chksum, y, w, fd_counter);
		}

	putst2(Display_UART2_msg);					// Say hello to UART2

// RS-232 I/O Test
	if(_DISPLAY) rs232_test();					// If CONFIG/DISPLAY pressed during startup Goto RS-232 test loop

// Clear Yaesu CAT buffer
	clear_cat();

// Clear Kenwood CAT receive buffer
	clr_rx_buffer();

// Clear TRX-2 - PA-100D message receiver buffer
	clear_buffer();
	auto_inform = OFF;							// Set Auto-Inform OFF
	reset_state();								// Reset Auto-Inform datapoints

// Set amateur band frequency limits
	clear_lcd();
	lcd_putst(band_limits[cal.calval.Band_Limits]);
	ms_delay(1000);

	if(cal.calval.Band_Limits)					// First set the upper frequency limits for both US licence classes
		{
		upper_limit[2] = _80M_UPPER_USA;
		upper_limit[3] = _40M_UPPER_USA;
		lower_limit[1] = 1800000L;				// US has the use of the bottom 10kHz of 160m.
		}

	if(cal.calval.Band_Limits == USA_GENERAL)	// Now set the lower frequency limits for the US General class
		{
		lower_limit[2] = 3525000L;
		lower_limit[3] = 7025000L;
		lower_limit[5] = 14025000L;
		lower_limit[7] = 21025000L;
		}

	if(cal.calval.alarm_flags & 0x08)			// If Under-Voltage alarm is enabled, set the low voltage warning limit
		low_voltage_warning = (int)((unsigned long)(((unsigned long)cal.calval.undervoltage_trip * (unsigned long)cal.calval.batt_mult) + 100000UL) / cal.calval.batt_mult);

	set_chgen(eeprom.defval.Display_Type);		// Load bar graph fonts

	if((display_page & 0x07) == LAST_PAGE)
		sub_page0 = (display_page >> 4);		// Restore last saved page
	else
		sub_page0 = display_page & 0x07;		// Restore preset page

	s_meter_timer = eeprom.defval.Hold_Time;
	old_cw_pot = 0x7FFF;						// Ensure CW Speed is set first time round
	A_equals_B_flag = FALSE;
	adjust_vfo();								// Ensure VFO frequency is adjusted to the frequency step size in use
	set_tx_filter_bw();
	power_bar_graph_const = (double)cal.calval.fwd_pwr_mult / 205417200.0;	// See notes in display_pwr_meter()
	id_const = int_round((long)cal.calval.overcurrent_trip, 48L);
	batt_mean_val = (double)convert_adc12(BATT_CH) * (double)(cal.calval.batt_mult);
	batt_scale_factor = cal.calval.overvoltage_trip - cal.calval.undervoltage_trip;

	if(cal.calval.swr_delay)
		loop_delay_time = int_round((long)(cal.calval.swr_delay) * 10000L, LOOP_DELAY);
	else
		loop_delay_time = SWR_AL_DLY;

	RS_232_Mode = port_mode[eeprom.defval.serial_mode];	// Set current Serial Port Mode (Avoids pointless loop invariant address calculation)

// Main forever loop _____________________________________________________________________________________________________________

	do	{
		vfo_lock = check_bit(vfo_idx);				// Set/Clear VFO Lock for this VFO
		band_in_use = check_freq(freq_in_use);		// Check valid amateur band frequency, 1 - 9 (160m - 10m)
		split_flag = ((eeprom.defval.max_vfo == SPLIT_MODE) && (eeprom.defval.vfo_sel == SPLIT_MODE));
		tx_enable = (!eeprom.defval.OOB_Check || band_in_use);
		rapid_band_switch_flag = (PTT_IN && (split_flag == FALSE) && (lcd_mode == NORMAL_DISPLAY));
		check_alarm();

		if(bl_off_flag)								// Check LCD back light status
			{
			bl_off_flag = FALSE;
			bl_off();
			}

		lcd_cmd(LINE1);								// Start from line 1,
		lcd_dsp[lcd_mode]();						// and display current LCD major page mode data

/*
The following module causes the mode or filter change to blink twice, no matter what display page is
currently in use. The variable: mode_ctr gets set to MODE_COUNTER in the function: set_mode_blink(), and this
is defined as:(4 * MODE_CHANGE) - 1, MODE_CHANGE is defined as a value of 40, so mode_ctr gets set to 159.
If the mode_ctr variable has decremented to 1, then this causes the sub-page to be reset to it initial value.
The variable: sub_page0 is normally 0, to display the main page with the S-meter and mode/filter settings, but
pressing the DISPLAY/CONFIG button increments it to 1 to display S-meter + Wattmeter, 2 to display S-meter + SWR,
3 to display S-meter + Voltmeter, and 4 to display S-meter + Ammeter. So, if we are displaying any page other
than 0 when we change the mode or filter setting, we save sub_page0, set sub_page0 to 0 and display the mode/filter
settings for two blinks, then restore the original sub_page value.

The variable: dsp1_ctr is the counter for the mode change. The variable dsp2_ctr is the counter for the filter change.
Initially these variables get set by set_mode_blink() to 3, which causes a blank to be displayed in function: dm_0().
On the first time through this ensures that they get set also to 3, because 159 / 40 is 3 in integer arithmetic.
This value persistes until the mode_ctr has decremented to 119, whereupon 119 / 40 = 2. At this point in the
function: dm_0() we now display the message. When mode_ctr has decremented to 79, dsp_1_ctr gets set to 1, which
causes a blank to be displayed again. When mode_ctr has decremented to 39 dsp_1 is 0, and this again causes the
message to be displayed. Finally, mode_ctr decrements to 1, the original display page is restored, mode_ctr decrements
to 0, and we exit, and it's all over. A similar set of operations for dsp2_ctr for the filter display.
As the main loop takes about 8mS to execute, this gives roughly 400mSec on and 400mSec off.
*/
		if(mode_ctr)
			{
			if(mode_ctr == 1)
				sub_page0 = tmp_sub_page0;				// Restore previous display page

			if(dsp1_ctr)								// Mode change blink
				dsp1_ctr = mode_ctr / MODE_CHANGE;

			if(dsp2_ctr)								// Filter change blink
				dsp2_ctr = mode_ctr / MODE_CHANGE;

			mode_ctr--;									// Decrement the blink counter
			}
/*
  This test code is only used to measure the loop cycle time. Version 1.07k takes 7 seconds
  between beeps, giving a loop cycle time of 7mSec.
  Version 1.07n takes 14mSec when displaying frequencies to 1Hz.
  When using the "long" method with repeated division and multiplication the loop time is 6.3mSec,
  When displaying to 10Hz using sprintf(), the loop time is 7mSec.
  Version 1.07t loops in 7.8mSec
  Version 1.07w Build 17 loops in 7.9mSec (10 loops in 79 seconds)
  Version 1.08a Build 4 loops in 6.6mSec (10 loops in 66 seconds)
  Version 1.09a Build 4 loops in 8.85mSec (20 loops in 177 seconds)
  Version 1.09a Build 5 loops in 8.85mSec (20 loops in 177 seconds)
  See notes in lcd-trx2.c for further information
  Comment out the following 7 lines when not needed - 5B4AIY
*/
/*
		if(!loop_ctr2)
			{
			loop_ctr2 = 1000;
			beep(4700, SHORT_BEEP);
			}

		loop_ctr2--;
*/
		if(!tune_on) audio_src = eeprom.defval.af_input;	// Save current audio input state

		if(c && SW5)							// Copy source frequency to destination when SW5 is released and c is set
			{
			eeprom.defval.vfo[dst_vfo] = eeprom.defval.vfo[src_vfo];
			eeprom.defval.mode[dst_vfo] = eeprom.defval.mode[src_vfo];

			if(c == RIGHT_ARROW) changed_flag = TRUE;

			c = FALSE;							// Clear copy flag
			}

// Make atomic copy of PTT_IN signal for this processing loop. PTT_IN: 1 = RX, 0 = TX
		if(PTT_IN_IO)							// RX mode
			PTT_IN = RECEIVE;					// Copy I/O bit to status flag - RX
		else									// TX mode
			{
			if(!tx_enable)
				{
				bl_on();
				PTT_IN = RECEIVE;				// Force to RX state
				lcd_cmd(LINE1);					// Start from line 1
				lcd_putst(out_of_band);
/*
  This logic fixes the bug that allowed transmission in SSB even when out-of-band. The hardware logic is such
  that the PTT switch is connected directly to the TX/RX logic, as the PTT signal on the main board, sheet 3 of 3,
  REGULATORS/CONTROL/CONNECTORS, top left-hand corner, R71/C71 connected to IC4A. However, if the state is set to
  TUNE or CW, then the SSB/!CW is asserted, and this asserts the !TX-SSB signal. This is fed to the microphone
  amplifier on the main board, sheet 2 of 3, TX SECTION, bottom right-hand section, gate of TR5. This ensures that
  the microphone amplifier feedback loop is shorted, giving a large amount of attenuation, and ensuring that little
  or no RF is generated, even though the PTT switch is closed. Note that there may be a brief pulse of RF generated.
  If the PTT switch is closed immediately after this point has been passed, and there is some audio present, then
  RF will be generated until the loop cycles back around to here, whereupon it will be suppressed. It takes about
  6 to 8 mSec to cycle through the main loop, and thus in theory you could generate a brief pulse of RF.
  The state decoder is set to the correct machine state later on in this code - line 4166.
  Adrian Ryan - 5B4AIY 23/JUN/2012
*/
				if(eeprom.defval.mode[vfo_idx] < CW)		// Is the mode LSB or USB?
					{
					int mode;								// Temporary local variable.

					mode = eeprom.defval.mode[vfo_idx];		// Yes, so save the current mode,
					eeprom.defval.mode[vfo_idx] = TUNE;		// and set the mode temporarily to TUNE,
					send_packet_to_74hc595();				// and switch the hardware to the TUNE state,
					eeprom.defval.mode[vfo_idx] = mode;		// and now restore the original mode.
					}

				while(!PTT_IN_IO);					// Wait for PTT switch to be released,
				}
			else
				{
				PTT_IN = TRANSMIT;					// TX state
				bl_on();							// Set LCD back light
				}
			}

// VFO frequency in use select logic, A/B, Split, RX/TX
		vfo_idx = eeprom.defval.vfo_sel;			// Copy VFO selection for processing, adjust selected VFO if Split

// Evaluate CW keyer enable condition, Split mode needs special conditions
// Using the split_flag instead of the tortuous previous logic eliminated 12 lines of code.
		if(split_flag)								// If Split Mode is active,
			{
			vfo_idx = (PTT_IN) ? VFO_A : VFO_B;		// then use VFO-A for RX and VFO-B for TX.
			cw_active = (eeprom.defval.mode[VFO_B] == CW || eeprom.defval.mode[VFO_B] == CWR);
			}
		else
			cw_active = (eeprom.defval.mode[vfo_idx] == CW || eeprom.defval.mode[vfo_idx] == CWR);

// Tune function
		if(eeprom.defval.mode[vfo_idx] == TUNE)		// TUNE mode selected
			{
			if(!PTT_IN && tx_enable)				// KEY follows PTT switch in tune mode
				{
				KEY = ON;

				if(tune_on == OFF)
					{
					tune_on = ON;					// Set tune state
					eeprom.defval.af_input = LINE;
					beep(HZ698_45, cal.calval.beep_len); // Short 700Hz beep, enter tune state
					}
				}
			else
				{
				KEY = OFF;							// Remove carrier
				tune_on = OFF;						// Remove tune state
				eeprom.defval.af_input = audio_src;	// Restore previous state
				}
			}
		else										// Other modes
			{
			if(tune_on)								// Tune state should not be on in other modes
				{
				tune_on = OFF;						// Remove tune state
				KEY = OFF;							// Remove CW carrier
				} 
			}

// Handle rotary encoder ( VFO )
		if(enc_event && (lcd_mode == NORMAL_DISPLAY))// Adjust frequency if encoder rotated and disp page 0 selected
			{
			enc_event = 0;

			if(!vfo_lock)							// Not locked
				{
				bl_on();							// LCD back light
				eeprom.defval.vfo[vfo_idx] += (encoder_get() * f_steps[eeprom.defval.step_idx]);

				if(eeprom.defval.vfo[vfo_idx] > MAX_DDS) eeprom.defval.vfo[vfo_idx] = MAX_DDS;
				if(eeprom.defval.vfo[vfo_idx] < MIN_DDS) eeprom.defval.vfo[vfo_idx] = MIN_DDS;
				}
			else
				reset_encoder();
			}

// Select & calculate frequency in use
		freq_in_use = eeprom.defval.vfo[vfo_idx];								// Load frequency from selected VFO
		offset = 0;																// LSB/USB/TUNE do not need any offset

		if(PTT_IN)																// RECEIVE
			{
			set_scaf_clock(eeprom.defval.scaf_filters[eeprom.defval.filter]);	// Select Receive mode filter
			swr = 0;															// Reset SWR

			if(eeprom.defval.mode[vfo_idx] == TUNE)								// If we are in the TUNE mode and the extended delay is ON,
				swr_al_dly = loop_delay_time;									// then use a longer delay to allow tuning, (6.9mSec/loop)
			else
				swr_al_dly = SWR_AL_DLY;										// otherwise use the normal 500mS delay.

			rit = get_rit();													// rit is used elsewhere in the serial-trx2.c module

			if(eeprom.defval.rit_on)
				freq_in_use += (long)rit;										// Add RIT (RIT = about +/-1kHz)

//			offset = (eeprom.defval.mode[vfo_idx] == CW) ? -eeprom.defval.cw_shift : (eeprom.defval.mode[vfo_idx] == CWR) ? eeprom.defval.cw_shift : 0;
// Although the previous line is the elegant way to do this, it actually takes more than 30 bytes of additional code! A.Ryan - 5B4AIY 09/JUL/12

			if(eeprom.defval.mode[vfo_idx] == CW) offset = -eeprom.defval.cw_shift;
			if(eeprom.defval.mode[vfo_idx] == CWR) offset = eeprom.defval.cw_shift;
			}
		else												// TRANSMIT
			{												// Select Transmit mode filter
			if(eeprom.defval.mode[vfo_idx] < CW)			// SSB TX audio filter
				set_scaf_clock(eeprom.defval.scaf_filters[TX_FILTER]);	// Set TX  filter from setup struct
			else
				set_scaf_clock(SIDE_TONE_SCAF);				// Set filter for side tone when CW or Tune

			get_pwr();										// Convert forward & reverse A/D channels
			swr = calc_swr(fwd_adval, rev_adval);			// Calculate SWR

			if(swr_al_dly)									// PTT to SWR alarm active delay
				swr_al_dly--;								// Delay needed for linear amplifier switching defeat

// If SWR > trip limit, and SWR delay has timed out, and the SWR alarm flag has not been set
			if((swr > cal.calval.swr_trip) && (swr_al_dly == 0) && !(alarm_flag & 0x02))
				alarm_flag |= 0x02;							// Set SWR alarm flag
			}

// Set CW period
// Convert CW pot
		cw_pot = convert_adc12(CW_POT);						// cw_pot value is used also for squelch

		if(cw_pot != old_cw_pot)
			{
			if(eeprom.defval.sq)							// If pot function is Squelch
				cw_period = cw_speed(eeprom.defval.CW_Speed & UNLOCKED);
			else											// If pot function is CW Speed
				cw_period = cw_speed(cw_pot);				// cw_period time is in mS

			old_cw_pot = cw_pot;
			}

		if(eeprom.defval.keyer == STRAIGHT_KEY)				// If the keyer mode is STRAIGHT KEY,
			{
			if(eeprom.defval.CW_Release > MIN_CW_RELEASE)	// and the release time is set,
				cw_period = eeprom.defval.CW_Release;		// then CW Release time is the user setting.
			else
				cw_period = 7 * cw_speed(old_cw_pot);		// Else it is the default speed release time.
			}

// Set other keyer parameters
		keyer_mode = eeprom.defval.keyer;					// Copy keyer operating mode from EEPROM data

// Squelch code
		if((eeprom.defval.sq) && (PTT_IN))					// Run squelch when squelch ON and RX
			{
			if((s_meter + 2) > (cw_pot >> 6) && (mute == AUDIO_ON))	// Check biased meter value against set value, check mute status
				{
				start_scaf_clock();							// Audio on
				sq_timer = SQ_HOLD;							// Reload squelch open timer, 850mS
				squelch_on = OFF;

				if(eeprom.defval.sq == 2) bl_on();			// If Squelch-B is selected, turn backlighting on
				}
			else
				{
				if(sq_timer == 0)
					{
					stop_scaf_clock();						// Audio off
					squelch_on = ON;						// Squelch closed flag for Yaesu CAT interface
					}
				}
			}
		else												// Squelch not in use
			{
			squelch_on = OFF;

			if(mute) stop_scaf_clock();						// Audio OFF if mute flag set from test code
			else start_scaf_clock();						// SCAF clock ON if TX or squelch not in use
			}

// Read switches
// DISPLAY/CONFIG
		if(_DISPLAY)
			{
			bl_on();										// LCD back light
			long_push = LONG_PUSH;							// Set long push timer

			if((lcd_mode == NORMAL_DISPLAY) && (sub_page0 == MAX_SUB_PAGE0))
				beep(HZ466_85, cal.calval.beep_len);		// B-flat tone for basic display
			else
				{
				if((lcd_mode == CONFIG_DISPLAY) && (sub_page1 == MAX_SUB_PAGE1))
					beep(HZ466_85, cal.calval.beep_len);	// B-flat tone for basic display
				else
					beep(HZ587_31, cal.calval.beep_len);	// D tone beep for other displays
				}

			ms_delay(BUTTON_DEBOUNCE);						// Debounce delay place changed 03.08.2008

			while(_DISPLAY)
				{
				if(long_push == 0)
					{
					beep(HZ466_85, LONG_BEEP);				// Long beep for major page change
					break;
					}
				}

			if(long_push == 0)								// Major mode change
				{
				lcd_mode++;

				if(lcd_mode > MAX_LCD_MODE)
					{
					save_settings(0, 2);
					lcd_mode = NORMAL_DISPLAY;
					}
				}
			else
				{
				if(lcd_mode == NORMAL_DISPLAY)				// Select sub page to adjust
					{
					sub_page0++;
					sub_page0 %= (LAST_PAGE);
					}
				else
					{
					sub_page1++;
					sub_page1 %= (MAX_SUB_PAGE1 + 1);
					}
				}

			while(_DISPLAY);								// Wait until button released

			ms_delay(BUTTON_DEBOUNCE);
			reset_encoder();								// Reset encoder after display change
			}

// FAST/VFAST
		if(_VFAST && PTT_IN && (lcd_mode == NORMAL_DISPLAY))// Allow Tuning Rate changes only in Receive and not in Setup Mode
			{
			bl_on();										// LCD back light
			long_push = LONG_PUSH;							// Set long push timer

			if(eeprom.defval.step_idx == KILO_HZ)			// When changing from Very Fast tuning rate,
				eeprom.defval.step_idx = TEN_HZ;			// default to Medium, 10Hz
			else
				{
				eeprom.defval.step_idx++;					// Increment the tuning rate,
				eeprom.defval.step_idx %= 3;				// and keep within normal range, 0 - 2
				}

			beep(filter_beep[eeprom.defval.step_idx], cal.calval.beep_len);
			ms_delay(BUTTON_DEBOUNCE);						// Button make debounce

			while(_VFAST)
				{
				if(long_push == 0)
					{
					beep(HZ698_45, cal.calval.beep_len * 10);	// Long beep, Very Fast tuning
					eeprom.defval.step_idx = KILO_HZ;		// 1kHz steps
					break;
					}
				}

			while(_VFAST);									// Wait for button release

			ms_delay(BUTTON_DEBOUNCE);						// Button release debounce
			}

		adjust_vfo();										// Round VFO frequency to current step size, if enabled.

// RIT
		if(_RIT && PTT_IN && (lcd_mode == NORMAL_DISPLAY))	// Only allow RIT to be selected in RX mode, and not in Setup Mode
			{
			int rit_flag = TRUE;							// Temporary flag, local to this section.

			bl_on();										// LCD back light
			glitch_time = 50;								// Set longer glitch time for Band changes - 5B4AIY
			long_push = LONG_PUSH;							// Set long push timer

// Do not permit rapid band switching in TX mode, split mode, service mode or setup mode
			while(_RIT && rapid_band_switch_flag)
				{
				if(long_push == 0)
					{
					beep(HZ466_85, cal.calval.beep_len * 10);	// Long beep for band switch
					latch = TRUE;								// Set the page hold latch
					rit_flag = FALSE;

					while(_RIT);								// Wait for button release

					ms_delay(BUTTON_DEBOUNCE);
					break;
					}
				}

			reset_encoder();									// Avoids spurious band change if VFO knob rotated while pressing RIT button.

			while(latch && rapid_band_switch_flag)
				{
				if(long_push == 0)
					{
					if(vfo_lock)
						sprintf(lcdpbuff, vfo_locked, vfo_idx + 'A');
					else
						{
						band_in_use = select_nearest_band(eeprom.defval.vfo[vfo_idx]);
						sprintf(lcdpbuff, selected_band, band[band_in_use]);
						band_in_use += encoder_get_coarse(ENCODER_MEDIUM);
						max_min(&band_in_use, HIGHEST_BAND, LOWEST_BAND);

						if(eeprom.defval.Band_SW == _DEFAULT)// Use Default frequencies
							{
							eeprom.defval.vfo[vfo_idx] = amateur_bands[band_in_use - 1];

							if(eeprom.defval.Auto_SB)
								{
								if(band_in_use > 3) eeprom.defval.mode[vfo_idx] = USB;
								else eeprom.defval.mode[vfo_idx] = LSB;
								}
							}	
						else								// Use User Mode frequencies
							{
							eeprom.defval.vfo[vfo_idx] = eeprom.defval.vfo[band_in_use + 29];
							eeprom.defval.mode[vfo_idx] = eeprom.defval.mode[band_in_use + 29];
							}

						if(eeprom.defval.max_vfo == SPLIT_MODE)	// Prevent cross-band split frequency operation
							{								// Check to see if VFO-A and VFO-B are using the same band.
							if(check_freq(eeprom.defval.vfo[VFO_A]) != check_freq(eeprom.defval.vfo[VFO_B]))
								{							// The VFOs are tuned to different bands, so,
								if(vfo_idx == VFO_A)		// if VFO-A is the active VFO,
									{						// then VFO-B is not on the same band as VFO-A.
									eeprom.defval.vfo[VFO_B] = eeprom.defval.vfo[VFO_A];
									eeprom.defval.mode[VFO_B] = eeprom.defval.mode[VFO_A];
									}
								else						// Else VFO-B is the active VFO,
									{						// and VFO-A is not on the same band as VFO-B.
									eeprom.defval.vfo[VFO_A] = eeprom.defval.vfo[VFO_B];
									eeprom.defval.mode[VFO_A] = eeprom.defval.mode[VFO_B];			
									}
								}
							}
						}
					}

				lcd_cmd(LINE1);
				lcd_putst(lcdpbuff);
				lcd_cmd(LINE2);
				disp_freq(eeprom.defval.vfo[vfo_idx]);
				lcd_putst(band_sw_msg[eeprom.defval.Band_SW]);
				warning_beep();

				if((PWR_SW || _RIT) && latch)
					{
					latch = FALSE;							// Release the page hold latch
					beep(HZ587_31, cal.calval.beep_len);
					}
				}

			if(rit_flag)
				{
				if(eeprom.defval.rit_on)
					{
					beep(HZ466_85, cal.calval.beep_len);	// B-flat tone beep for RIT off
					eeprom.defval.rit_on = OFF;
					}
				else
					{
					beep(HZ587_31, cal.calval.beep_len);	// D tone beep for RIT on
					eeprom.defval.rit_on = ON;
					}
				}

			while(PWR_SW || _RIT);							// Wait for button to be released

			ms_delay(BUTTON_DEBOUNCE);
			reset_encoder();								// Reset encoder after band change
			}

// VFO
		if(_VFO && PTT_IN && (lcd_mode == NORMAL_DISPLAY))	// Do not change if TX is ON, or if in User Setup Mode!
			{
			int encoder, selected_vfo, not_moved = TRUE, single_step = TRUE;	// Temporary local variables and flags, local to this section.

			bl_on();										// LCD back light
			glitch_time = 100;								// Set longer glitch time for VFO changes - 5B4AIY
			selected_vfo = eeprom.defval.vfo_sel;
			long_push = LONG_PUSH;

			if(eeprom.defval.max_vfo == SPLIT_MODE)			// A/B + Split Mode
				{
				if(eeprom.defval.vfo_sel == SPLIT_MODE) beep(HZ466_85, cal.calval.beep_len);	// B-flat tone for VFO-A
				else beep(HZ587_31, cal.calval.beep_len);	// D tone for VFO-B & Split

				eeprom.defval.mode[SPLIT_MODE] = LSB;		// Bug fix, Split mode phantom VFO #2 should never be in Tune mode (Mode = 3), 03.07.2008

				while(_VFO)
					{
					if(long_push == 0)
						{
						beep(HZ466_85, cal.calval.beep_len * 10);	// Long beep for VFO A to B copy
						break;
						}
					}	

				if(long_push == 0)							// VFO copy A to B or B to A
					{

					if(vfo_idx == VFO_A)					// If VFO A is in use
						{
						if(check_bit(VFO_B))
							sprintf(lcdpbuff, vfo_locked, 'B');
						else
							{
							eeprom.defval.vfo[VFO_B] = eeprom.defval.vfo[VFO_A];	// Copy A to B
							eeprom.defval.mode[VFO_B] = eeprom.defval.mode[VFO_A];	// Copy also mode
							sprintf(lcdpbuff, copy_vfo, 'A', RIGHT_ARROW, 'B');
							A_equals_B_flag = 2;
							}
						}
					else									// If VFO B is in use
						{
						if(check_bit(VFO_A))
							sprintf(lcdpbuff, vfo_locked, 'A');
						else
							{
							eeprom.defval.vfo[VFO_A] = eeprom.defval.vfo[VFO_B];	// Copy B to A
							eeprom.defval.mode[VFO_A] = eeprom.defval.mode[VFO_B];	// Copy also mode
							sprintf(lcdpbuff, copy_vfo, 'B', RIGHT_ARROW, 'A');
							A_equals_B_flag = 1;
							}
						}

					lcd_cmd(LINE2);							// Second display line
					lcd_putst(lcdpbuff);
					ms_delay(BUTTON_DEBOUNCE);				// Debounce delay added 03.08.2008

					while(_VFO);							// Show while button pressed

					reset_encoder();
					}
				else
					{
					eeprom.defval.vfo_sel++;
					eeprom.defval.vfo_sel %= 3;				// Alter VFO A, B, Split
					}
				}
			else											// Linear memory bank mode
				{
/*
  VFO Select Bug
  When the VFO Multi-Memory option was selected, if the VFO button was held down to permit rapid selection of
  an active VFO by means of the VFO knob, if the last VFO was selected, then when the VFO button was next held
  down for a long time the first VFO knob movement would either select VFOB if it was rotated clockwise, or
  VFOA if rotated anticlockwise instead of either staying at the maximum selected VFO or decrementing it.
  This occured because of a spurious increment in the previous lines of code, and the resetting of the VFO
  number to zero if the count exceeded the maximum. Since the variable: vfo_idx always gets set to the same
  as eeprom.defval.vfo_sel, by temporarily using vfo_idx since it will not have been changed, it is possible
  to overcome this bug. Note that it is the vfo_idx variable that is incremented, and its final value assigned
  to eeprom.defval.vfo_sel. If instead the vfo_idx contents were simply copied to vfo_sel, then as it would not
  change, the vfo_sel variable would only increment once. Equally, if the vfo_idx variable is incremented and
  then copied to vfo_sel, although this will work perfectly well, there is a hidden flaw. Since the checks
  are only performed on the vfo_sel variable, once it has been set to its maximum value the vfo_idx would
  continue to increment if the VFO knob were continually rotated, and eventually, after a very large number
  of rotations, since the vfo_idx variable is a signed integer, it would suddenly overflow negative, and the
  vfo_sel variable would be set to zero. By using the vfo_idx variable throughout the increments and decrements
  and performing the checks on this variable and then finally copying it to vfo_sel, this bug is avoided, and
  at the same time vfo_idx and vfo_sel are synchronised, which is also a good feature. - Adrian Ryan 20/NOV/2011
*/
				while(_VFO)					// Use VFO knob to rapidly select the desired active VFO
					{
					if(long_push == 0)
						{
						beep(HZ466_85, cal.calval.beep_len * 10);	// Long beep for latched mode
						latch = TRUE;								// Set the page hold latch
						single_step = FALSE;

						while(_VFO);								// Wait for button release

						ms_delay(BUTTON_DEBOUNCE);
						break;
						}
					}

				while(latch)
					{
					encoder = encoder_get_coarse(ENCODER_MEDIUM);	// Read encoder, 26 memories = about 1 turn

					if(encoder != 0)								// Encoder moved
						{
						not_moved = FALSE;
						vfo_idx += encoder;						// This fixes the VFO Select Bug!

						if(vfo_idx >= eeprom.defval.max_vfo) vfo_idx = eeprom.defval.max_vfo - 1;// Check upper limit
						if(vfo_idx < VFO_A) vfo_idx = VFO_A;	// Check lower limit

						eeprom.defval.vfo_sel = vfo_idx;
// Setting the vfo_lock here avoids a spurious blink of the lock symbol when releasing the VFO button
						vfo_lock = check_bit(eeprom.defval.vfo_sel);
						lcd_cmd(MODE_FIELD);					// Move cursor to Line-1 Mode display field
						lcd_putst(mode_msg[eeprom.defval.mode[eeprom.defval.vfo_sel]]);	// Display the new mode
						lcd_cmd(LINE2);							// Second display line
						disp_freq(eeprom.defval.vfo[eeprom.defval.vfo_sel]);			// Display new frequency
						sprintf(lcdpbuff, Display_VFO_Lock, eeprom.defval.vfo_sel + 'A', lock_symbol[vfo_lock]);
						lcd_putst(lcdpbuff);					// Display selected VFO, Lock/Unlock status, and current tuning rate
						}

					warning_beep();

					if((PWR_SW || _VFO) && latch)
						{
						latch = FALSE;							// Release the page hold latch
						beep(HZ587_31, cal.calval.beep_len);

						if(not_moved) eeprom.defval.vfo_sel = selected_vfo;		// Fix spurious VFO Select if VFO knob is not moved.
						}
					}

				if(single_step)
					{
					if(eeprom.defval.vfo_sel == (eeprom.defval.max_vfo - 1)) beep(HZ466_85, cal.calval.beep_len);	// B-flat tone for VFO change
					else beep(HZ587_31, cal.calval.beep_len);	// D tone for first location

					eeprom.defval.vfo_sel++;
					eeprom.defval.vfo_sel %= eeprom.defval.max_vfo;
					}
				}

			while(PWR_SW || _VFO);

			ms_delay(BUTTON_DEBOUNCE);
			reset_encoder();								// Reset encoder on exit
			}

// MODE
		if(_MODE && PTT_IN)									// Do not change if TX is ON !
			{
			int mode_change = TRUE;							// Temporary flag, local to this section.
			int vfo = 0;									// Temporary variable, local to this section.

			bl_on();										// LCD back light
			glitch_time = 5;								// Set glitch timer for short time
			long_push = LONG_PUSH;

			while(_MODE && (split_flag == FALSE))
				{
				if(long_push == 0)
					{
					if(lcd_mode == CONFIG_DISPLAY)
						beep(HZ587_31, LONG_BEEP);			// Long beep for Fast Config Menu Page Switch
					else
						beep(HZ466_85, cal.calval.beep_len * 10);	// Long beep for copy to/from VFO memories

					reset_encoder();						// Finally! This fixes the spurious page change when using the Fast Page Switch.
					latch = TRUE;
					mode_change = FALSE;

					while(_MODE);							// Wait for button release

					ms_delay(BUTTON_DEBOUNCE);
					break;
					}
				}

			while(latch && (lcd_mode == CONFIG_DISPLAY))	// Fast User Config page select
				{
				if(long_push == 0)
					sub_page1 = display_settings(sub_page1, TRUE);

				warning_beep();

				if((PWR_SW || _MODE) && latch)
					{
					latch = FALSE;							// Release the page hold latch
					mode_change = FALSE;
					beep(HZ587_31, cal.calval.beep_len);
					}
				}

			while(latch && (lcd_mode == NORMAL_DISPLAY) && (split_flag == FALSE))	// VFO Multi-Copy Mode
				{
				vfo += encoder_get_coarse(ENCODER_MEDIUM);
				max_min(&vfo, 25, -26);

				if(vfo < 0)			// Copy memory to Active VFO
					{
					c = LEFT_ARROW;	// Left direction arrow
					src_vfo = -(vfo + 1);
					dst_vfo = eeprom.defval.vfo_sel;
					sprintf(lcdpbuff, copy_vfo, (char)(dst_vfo + 'A'), c, (char)(src_vfo + 'A'));
					}
				else				// Copy Active VFO to memory
					{
					c = RIGHT_ARROW;// Right direction arrow
					src_vfo = eeprom.defval.vfo_sel;
					dst_vfo = vfo;
					sprintf(lcdpbuff, copy_vfo, (char)(src_vfo + 'A'), c, (char)(dst_vfo + 'A'));
					}

				lcd_cmd(LINE1);
				lcd_putst(lcdpbuff);
				lcd_cmd(LINE2);
				disp_freq(eeprom.defval.vfo[src_vfo]);

				if(check_bit(dst_vfo))
					{
					lcd_putst(band_sw_msg[_LOCKED]);	// Message is 'Locked'
					c = FALSE;							// Clear copy flag
					}
				else
					{
					lcd_spc(3);

					if(eeprom.defval.mode[src_vfo] != TUNE)
						lcdoutch(' ');

					lcd_putst(mode_msg[eeprom.defval.mode[src_vfo]]);
					}

				mode_ctr = 1;
				dsp2_ctr = 0;							// Prevent spurious blink
				warning_beep();

				if((PWR_SW || _MODE) && latch)
					{
					latch = FALSE;						// Release the page hold latch
					mode_change = FALSE;
					beep(HZ587_31, cal.calval.beep_len);
					}
				}

			if(mode_change)
				{
				if(!mode_ctr) dsp2_ctr = set_mode_blink();

				if(lcd_mode == NORMAL_DISPLAY)					// Do not allow MODE changes in User Config
					{
					if(eeprom.defval.mode[vfo_idx] == TUNE)		// This push returns us to Lower Side Band (0)
						beep(HZ466_85, cal.calval.beep_len);	// B-flat tone for LSB
					else
						beep(HZ587_31, cal.calval.beep_len);	// D tone for others

					eeprom.defval.mode[vfo_idx]++; 				// Adjust mode, 0 = LSB, 1 = USB, 2 = CW, 3 = CWR, 4 = Tune
					eeprom.defval.mode[vfo_idx] %= 5;			// Keep within limits 0 - 4
					}
				}

			while(PWR_SW || _MODE);						// Wait until button is released

			ms_delay(BUTTON_DEBOUNCE);					// Release debounce
			reset_encoder();
			}

// FILTER
// Only permit filter changes in receive mode and in the user configuration mode when displaying the filter bandwidth
		if(_FILTER && PTT_IN && ((lcd_mode == NORMAL_DISPLAY) || (lcd_mode == CONFIG_DISPLAY && (sub_page1 == 1))))
			{
			int filter_change = TRUE;						// Temporary flag, local to this section.

			bl_on();
			long_push = LONG_PUSH;							// Set long push timer

			if((eeprom.defval.mode[vfo_idx] == CWR) && (eeprom.defval.CW_CWR_Swap == ON))
				{
				// If in CWR mode and CW/CWR Swap is On, go back to CW
				beep(HZ587_31, cal.calval.beep_len);		// Indicate we have seen the button pushed,

				eeprom.defval.mode[vfo_idx] = CW;			// reset the mode to CW,

				if (!mode_ctr) dsp2_ctr = set_mode_blink();	// Initialise the mode blink counter.

				filter_change = FALSE;
				}
			else
				{
// Do not permit storage of user frequency when in TX, split, service or setup modes.
				while(_FILTER && rapid_band_switch_flag)
					{
					if(long_push == 0)
						{
						if(eeprom.defval.Band_SW == _DEFAULT)	// If in the DEFAULT mode...
							{
							user_band_prompt = not_in_user_mode;
							beep(HZ698_45, cal.calval.beep_len);
							}
						else if(eeprom.defval.Band_SW == _USER)	// Or, if in the USER mode...
							{
							if(!band_in_use)
								{
								user_band_prompt = out_of_band;
								beep(HZ698_45, cal.calval.beep_len);
								}
							else
								{
								eeprom.defval.vfo[band_in_use + 29] = eeprom.defval.vfo[vfo_idx];
								eeprom.defval.mode[band_in_use + 29] = eeprom.defval.mode[vfo_idx];
								sprintf(lcdpbuff, save_to_band, band[band_in_use]);
								user_band_prompt = lcdpbuff;
								changed_flag = TRUE;
								beep(HZ466_85, cal.calval.beep_len * 10);
								}
							}
							else							// Otherwise, must be in the LOCKED mode.
								{
								user_band_prompt = user_mem_locked;
								beep(HZ698_45, cal.calval.beep_len);
								}

						lcd_cmd(LINE1);
						lcd_putst(user_band_prompt);
						mode_ctr = 1;
						dsp1_ctr = 0;						// Cancel the blink
						filter_change = FALSE;

						while(_FILTER);						// Wait while FILTER switch is pressed
						}
					}

				if(filter_change)
					{
					if(!mode_ctr) dsp1_ctr = set_mode_blink();

					eeprom.defval.filter++; 				// Adjust filter select, 0 = Nar, 1 = Mid, 2 = Wide, 3 = Tx

					if(lcd_mode & (sub_page1 == 1))			// If we are in the User Configuration Mode, TX filter
						eeprom.defval.filter %= 4;
					else									// Normal mode, RX filter
						eeprom.defval.filter %= 3;

					beep(filter_beep[eeprom.defval.filter], cal.calval.beep_len);
					}

				while(_FILTER);								// Wait until button is released

				ms_delay(BUTTON_DEBOUNCE);					// Button release debounce
				reset_encoder();
				}
			}
// Load synthesiser with new frequency
		dds_load(freq_in_use + offset);
// Build main board & RF filter board (Only All-Band version) 74HC595 I/O latch data & send it
		send_packet_to_74hc595();
// Run UART1 Serial Interface functions
		RS_232_Mode();
// Run UART2 / AUX connector serial interface functions. (Voice Memory Control)
		uart2_ext_kb();
// Test shut down
		if(PWR_SW)
			{
			bl_on();										// Set LCD back light
			power_off();									// Go to shut down or toggle Lock or cancel save operation
			}
		} while(TRUE);	// End infinite DO-loop
	}	// End main

