/*
 The button scan function responds to the following buttons:
 SW6	FILTER	0 = Pressed
 SW5	MODE	0 = Pressed
 SW4	VFO		0 = Pressed
 SW3	RIT		0 = Pressed
 SW2	VFAST	0 = Pressed
 SW1	DISPLAY	0 = Pressed

 _FILTER		1 = Pressed
 _MODE			1 = Pressed
 _VFO			1 = Pressed
 _RIT			1 = Pressed
 _VFAST			1 = Pressed
 _DISPLAY		1 = Pressed
 _PWR_SW		1 = Pressed

 The function returns a function execution index, used to initiate one of the button functions.
 This leaves the main loop free for more important activities.
 In the idle state, with no buttons pressed, the function merely returns a scan code of zero, which does nothing.
 From the idle state, (idle flag == 0) if a button is pressed, this starts the button timer. A non-zero value in the button timer causes
 the main 1mS timer interrupt to increment the timer. If the timer should exceed a limit value, then it is pegged at that
 value. With a zero function code the buttons are sequenctially scanned, and if one is pressed this starts the button timer, and sets the
 function code to a non-zero number representing the function that has been initially selected. The function returns zero at this point
 awaiting a button release. When the button is released the button timer value is examined, and if it is less than 500mS this is a short
 push, and the appropriate function index is returned, and the button timer is reset to zero. The return value is used as an index into a
 table of functions, and when the function terminates the idle flag is set. This causes the function code, scan code, and button timers to be reset
 restoring the idle state once again.
*/
int set_scan_code(int function_code)
	{
	if(button_timer < 100) return 0;			// This is probably a button bounce event.
	if(button_timer > 500) function_code += 1;	// This is a more reasonable button release event.
	button_timer = 0;							// Reset the button timer,
	return function_code;						// and return the actual function execution index.
	}

int button_scan(void)
	{
	int static function_code;

	if(button_timer == 0)						// Normal idle state, scan buttons looking for a button pressed.
		{
		function_code = 0;						// Ensure that the function code is reset
		if(_FILTER) function_code = 1;
		if(_MODE) function_code = 2;
		if(VFO) function_code = 3
		if(_RIT) function_code = 4;
		if(_VFAST) function_code = 5;
		if(_DISPLAY) function_code = 6;
		if(_PWR_SW) function_code = 7;
		if(function_code) button_timer = 1;		// If a button was pressed, then start the button timer
		return 0;
		}
	else										// A button was pressed, scan for its release and return the function execution index.
		{
		switch(function_code)
			{
			case 0: default:
				return 0;
			break;

			case 1:			// FILTER
				if(SW6) function_code = set_scan_code(1);
			break;

			case 2:			// MODE
				if(SW5) function_code = set_scan_code(3);
			break;

			case 3:			// VFO
				if(SW4) function_code = set_scan_code(5);
			break;

			case 4:			// RIT
				if(SW3) function_code = set_scan_code(7);
			break;

			case 5:			// VFAST
				if(SW2) function_code = set_scan_code(9);
			break;

			case 6:			// DISPLAY
				if(SW1) function_code = set_scan_code(11);
			break;

			case 7:			// PWR
				if(!PWR_SW) function_code = set_scan_code(13);
			break;
			}
		}
	return function_code;
	}

